# import libraries
import json
from typing import Any, List, Optional

# import locals
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class EventList(Feature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._event_list = []

        # Map of event names to primary detail parameter and its type
        self._details_map = {
            "accept_job":                   ("job_name", "string_value"),
            "switch_job":                   ("job_name", "string_value"),
            "receive_fact":                 ("fact_id", "string_value"),
            "receive_entity":               ("entity_id", "string_value"),
            "complete_job":                 ("job_name", "string_value"),
            "complete_task":                ("task_id", "string_value"),
            "scene_changed":                ("scene_name", "string_value"),
            "room_changed":                 ("room_name", "string_value"),
            "begin_dive":                   ("site_id", "string_value"),
            "ask_for_help":                 ("node_id", "string_value"),
            "script_fired":                 ("node_id", "string_value"),
            "bestiary_select_species":      ("species_id", "string_value"),
            "bestiary_select_environment":  ("environment_id", "string_value"),
            "bestiary_select_model":        ("model_id", "string_value"),
            "begin_model":                  ("job_name", "string_value"),
            "model_phase_changed":          ("phase", "string_value"),
            "model_ecosystem_selected":     ("ecosystem", "string_value"),
            "model_concept_started":        ("ecosystem", "string_value"),
            "model_concept_updated":        ("status", "string_value"),
            "model_concept_exported":       ("ecosystem", "string_value"),
            "begin_simulation":             ("job_name", "string_value"),
            "model_sync_error":             ("sync", "int_value"),
            "simulation_sync_achieved":     ("job_name", "string_value"),
            "model_predict_completed":      ("ecosystem", "string_value"),
            "model_intervene_update":       ("difference_value", "int_value"),
            "model_intervene_error":        ("ecosystem", "string_value"),
            "model_intervene_completed":    ("ecosystem", "string_value"),
            "end_model":                    ("phase", "string_value"),
            "purchase_upgrade":             ("item_name", "string_value"),
            "insufficient_funds":           ("item_name", "string_value"),
            "add_environment":              ("environment", "string_value"),
            "remove_environment":           ("environment", "string_value"),
            "add_critter":                  ("critter", "string_value"),
            "remove_critter":               ("critter", "string_value"),
            "begin_experiment":             ("tank_type", "string_value"),
            "end_experiment":               ("tank_type", "string_value"),
            "begin_argument":               ("job_name", "string_value"),
            "fact_submitted":               ("fact_id", "string_value"),
            "fact_rejected":                ("fact_id", "string_value"),
            "leave_argument":               ("job_name", "string_value"),
            "complete_argument":            ("job_name", "string_value")
        }

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.UserID:
            next_event = {
                "name": event.EventName,
                "user_id": event.UserID,
                "session_id": event.SessionID,
                "timestamp": event.Timestamp.isoformat(),
                "job_name": event.GameState.get('job_name', event.EventData.get('job_name', "UNDEFINED")),
                "index": event.EventSequenceIndex,
                "event_primary_detail": None
            }

            if event.EventName == "scene_changed":
                next_event['scene_name'] = event.EventData['scene_name']

            if event.EventName in self._details_map:
                param_name = self._details_map[event.EventName][0]

                try:
                    next_event["event_primary_detail"] = event.EventData[param_name]
                except KeyError as err:
                    raise KeyError(f"Event of type {event.EventName} did not have parameter {param_name}, valid parameters are {event.EventData.keys()}")

            self._event_list.append(next_event)

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [json.dumps(self._event_list)]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



from typing import Any, Final, List
from enum import Enum
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class Thermotool(Enum):
    INSULATION          = "insulation"
    LOWER_STOP          = "lower_stop"
    UPPER_STOP          = "upper_stop"
    INCREASE_WEIGHT     = "increase_weight"
    DECREASE_WEIGHT     = "decrease_weight"
    HEAT                = "heat"
    COOLING             = "cooling"
    CHAMBER_TEMPERATURE = "chamber_temperature"
    CHAMBER_PRESSURE    = "chamber_pressure"

class ToolNudgeCount(SessionFeature):

    def __init__(self, params:GeneratorParameters, player_id:str):
        self._player_id = player_id
        self._tool_nudge_count = {tool: 0 for tool in Thermotool}
        super().__init__(params=params)

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["click_tool_increase", "click_tool_decrease"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
            if event.EventName == "click_tool_increase" or event.EventName == "click_tool_decrease":
                tool_name = event.EventData.get('tool_name', None)
                if tool_name:
                    tool = Thermotool(tool_name)
                    self._tool_nudge_count[tool] += 1

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return list(self._tool_nudge_count.values())



# import libraries
import json
from typing import Any, Dict, List, Optional
# import local files
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature




class FailureAttributes(PerLevelFeature):
    """Per-level feature to generate a list of failures, where each failure in the list is the set of player attributes

    :param PerLevelFeature: Base class for a Custom Feature class.
    :type PerLevelFeature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        PerLevelFeature.__init__(self, params=params)

        self._fails : List[Dict[str, int]] = []

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _getEventDependencies(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["level_fail"] # >>> fill in names of events this Feature should use for extraction. <<<

    @classmethod
    def _getFeatureDependencies(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of first-order features this Feature should use for extraction. <<<

    def _extractFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        attribs = ["endurance", "resourceful", "tech", "social", "trust", "research"]
        stats = json.loads(event.GameState.get("current_stats", str([None]*len(attribs))))
        self._fails.append(dict(zip(attribs, stats)))
        return

    def _extractFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        # >>> use data in the FeatureData object to update state variables as needed. <<<
        # Note: This function runs on data from each Feature whose name matches one of the strings returned by _getFeatureDependencies().
        #       The number of instances of each Feature may vary, depending on the configuration and the unit of analysis at which this CustomFeature is run.
        return []

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        return [self._fails]
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.POPULATION, ExtractionMode.PLAYER, ExtractionMode.SESSION] # >>> delete any modes you don't want run for your Feature. <<<



# import standard libraries
from datetime import datetime
from typing import Callable, List

# import local files
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.detectors.DetectorEvent import DetectorEvent
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode

class CollectFactNoJob(Detector):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters, trigger_callback:Callable[[Event], None]):
        super().__init__(params=params, trigger_callback=trigger_callback)
        self._found_jobless_fact = False
        self._fact_id = None
        self._current_job = None

    # *** Implement abstract functions ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["receive_fact"]

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        self._current_job = event.GameState.get('job_name', event.EventData.get('job_name', None))
        if self._current_job is None:
            raise KeyError("Could not find key 'job_name' in GameState or EventData!")
        if self._current_job == "no-active-job":
            self._found_jobless_fact = True
            self._fact_id = event.EventData['fact_id']
        return

    def _trigger_condition(self) -> bool:
        if self._found_jobless_fact:
            self._found_jobless_fact = False
            return True
        else:
            return False

    def _trigger_event(self) -> DetectorEvent:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        ret_val : DetectorEvent = self.GenerateEvent(app_id="AQUALAB",
                                                     event_name="CollectFactNoJob", event_data={"fact_id":self._fact_id},
                                                     game_state={"job_name":self._current_job})
        return ret_val



# import standard libraries
from typing import Any, Dict
# import local files
from ogd.core.schemas.configs.data_sources.DataSourceSchema import DataSourceSchema

class FileSourceSchema(DataSourceSchema):
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        super().__init__(name=name, other_elements=all_elements)

    @property
    def AsMarkdown(self) -> str:
        ret_val : str

        ret_val = f"FILE SOURCE"
        return ret_val

    @property
    def AsConnectionInfo(self) -> str:
        ret_val : str = f"{self._name}"
        return ret_val



from typing import Any, List
import json
import pandas as pd
import numpy as np
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event

orderMapping = {'1. One Box': 1, '2. Separated Boxes': 2, '3. Rotate a Pyramid': 3, '4. Match Silhouettes': 4, '5. Removing Objects': 5, '6. Stretch a Ramp': 6, '7. Max 2 Boxes': 7, '8. Combine 2 Ramps': 8, '9. Scaling Round Objects': 9,
                'Square Cross-Sections': 10, 'Bird Fez': 11, 'Pi Henge': 12, '45-Degree Rotations': 13,  'Pyramids are Strange': 14, 'Boxes Obscure Spheres': 15, 'Object Limits': 16, 'Warm Up': 17, 'Angled Silhouette': 18,
                'Sugar Cones': 19,'Stranger Shapes': 20, 'Tall and Small': 21, 'Ramp Up and Can It': 22, 'More Than Meets Your Eye': 23, 'Not Bird': 24, 'Unnecesary': 25, 'Zzz': 26, 'Bull Market': 27, 'Few Clues': 28, 'Orange Dance': 29, 'Bear Market': 30}
listPuzzles = list(orderMapping.keys())

eventsWithMetadata = ["create_shape", "delete_shape", "rotate_shape", "scale_shape", "move_shape"]
selectedEvents = ["start_level", "create_shape", "delete_shape", "rotate_shape", "scale_shape", "move_shape", "check_solution", "puzzle_complete", "rotate_view", "undo_action", "redo_action", "snapshot"]
#undoableActions = ["create_shape", "delete_shape", "rotate_shape", "scale_shape", "move_shape", "select_shape", "deselect_shape", #"select_shape_add", "rotate_view", "mode_change", "palette_change", "paint", "toogle_snapshot_display"]

class SequenceWithinPuzzles(SessionFeature):
    def __init__(self,  params:GeneratorParameters):
        super().__init__(params=params)
        self._numPuzzles = 1
        self._currentPuzzle = []
        self._activePuzzle = None
        self._prevEvent = None
        self._prevCheck = False
        self._dictFigures = {}
        self._userPuzzleDict = dict()

    def GetEventTypes(self) -> List[str]:
        return ["start_level", "puzzle_started", "create_shape", "delete_shape", "rotate_shape", "scale_shape", "move_shape", "check_solution", "puzzle_complete", "rotate_view", "undo_action", "redo_action", "snapshot", "disconnect","login_user", "exit_to_menu"]

    def GetFeatureValues(self) -> List[Any]:
        listReturn = []
        listReturn.append(self._numPuzzles - 1)
        emptyList = {}
        for puzzle in listPuzzles:
            if puzzle in self._userPuzzleDict.keys():
                listReturn.append(json.dumps(self._userPuzzleDict[puzzle]))
            else:
                listReturn.append(emptyList)
        return listReturn
        
    def Subfeatures(self) -> List[str]:
        return listPuzzles

    def _extractFromEvent(self, event:Event) -> None:
        appendEvent = False
        ignoreEvent = False
        currentEvent = {}
        currentEvent["type"] = event.event_name
        currentEvent["n_times"] = 1
        currentEvent["metadata"] = {}
        currentEvent["action_index"] = [np.nan]
        
        if (self._prevCheck == True):
            if (event.event_name == "puzzle_complete"):
                self._prevEvent["metadata"]["correct"] = True
            else:
                self._prevEvent["metadata"]["correct"] = False
            
            self._prevCheck = False
                
                
        if event.event_name == "start_level":
            self._activePuzzle = event.event_data["task_id"]["string_value"]
            
            if self._activePuzzle == "Sandbox":
                ignoreEvent = True
            
            if ignoreEvent == False:
                if self._activePuzzle not in self._userPuzzleDict.keys():
                    self._userPuzzleDict[self._activePuzzle] = {}
                #self._currentPuzzle = {}
                #self._currentPuzzle["sequence"] = self._numPuzzles
                #self._currentPuzzle["funnel"] = "No data"
        elif event.event_name not in ["login_user", "puzzle_started"]:
            currentEvent["action_index"] = [event.event_data["actionIndex"]["int_value"]]

        if self._activePuzzle == "Sandbox":
            ignoreEvent = True
            
        if ignoreEvent == False:
        
            if event.event_name == "check_solution":
                self._prevCheck = True
                appendEvent = False
            
            elif event.event_name == "create_shape":
                shape_id = event.event_data["objSerialization"]["int_value"]
                shape_type = event.event_data["shapeType"]["int_value"]
                self._dictFigures[shape_id] = shape_type
                currentEvent["metadata"]["shape_id"] = shape_id
                currentEvent["metadata"]["shape_type"] = shape_type
            
            # Should be allowed for one or more objects, but until now we will treat this case as one single object, int
            elif event.event_name in ["delete_shape", "move_shape"]:
                if event.event_name == "delete_shape":
                    shape_id = event.event_data["deletedShapes"]["int_value"]
                else:
                    shape_id = event.event_data["selectedObjects"]["int_value"]
                shape_type = self._dictFigures.get(shape_id)
                if shape_type == None:
                    self._dictFigures[shape_id] = shape_type
                currentEvent["metadata"]["shape_id"] = shape_id
                currentEvent["metadata"]["shape_type"] = shape_type
                
            elif event.event_name in ["rotate_shape", "scale_shape"]:
                shape_id = event.event_data["selectedObject"]["int_value"]
                shape_type = self._dictFigures.get(shape_id)
                if shape_type == None:
                    self._dictFigures[shape_id] = shape_type
                currentEvent["metadata"]["shape_id"] = shape_id
                currentEvent["metadata"]["shape_type"] = shape_type
            
            elif event.event_name in ["undo_action", "redo_action"]:
            
                currentEvent["metadata"]["targeted_actionIndex"] = event.event_data["targetedActionIndex"]["int_value"]
                
                
                    
            #if (appendEvent == True):
            #    self._currentPuzzle.append(json.dumps(currentEvent))
            
            if self._prevEvent != None:
                if event.event_name == self._prevEvent["type"]:
                    if event.event_name in eventsWithMetadata:
                        if currentEvent["metadata"]["shape_id"] != self._prevEvent["metadata"]["shape_id"]:
                            if (self._prevEvent["type"] in selectedEvents):
                                self._currentPuzzle.append(json.dumps(self._prevEvent))
                        else:
                            currentEvent["action_index"].extend(self._prevEvent["action_index"])
                            currentEvent["n_times"] = self._prevEvent["n_times"]
                            currentEvent["n_times"] += 1
                            
                    elif event.event_name in ["check_solution", "undo_action", "redo_action"]:
                        if (self._prevEvent["type"] in selectedEvents):
                            self._currentPuzzle.append(json.dumps(self._prevEvent))
                    else:
                        currentEvent["action_index"].extend(self._prevEvent["action_index"])
                        currentEvent["n_times"] = self._prevEvent["n_times"]
                        currentEvent["n_times"] += 1
                        
                else:
                    if (self._prevEvent["type"] in selectedEvents):
                        self._currentPuzzle.append(json.dumps(self._prevEvent))
            
            
            
            if event.event_name in ["disconnect", "login_user", "exit_to_menu"] and self._activePuzzle != None:
                #Add current data
                key = "Attempt" + str(self._numPuzzles)
                self._userPuzzleDict[self._activePuzzle][key] = self._currentPuzzle
                self._currentPuzzle = []
                self._dictFigures = {}
                self._numPuzzles += 1
            
            self._prevEvent = currentEvent




# import libraries
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData


class ClosenessIntercept(Feature):
    def __init__(self, params:GeneratorParameters):
        Feature.__init__(self, params=params)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
    #     closenesses = self.move_closenesses_tx[lvl]['completeness']
    #     times = self.move_closenesses_tx[lvl]['t']
    #     ranges = self.move_closenesses_tx[lvl]['range']
    #     if times:
    #         try:
    #             X = [(times[i]-times[0]).seconds for i in range(len(times))]
    #         except Exception as err:
    #             Logger.Log(times[0])
    #             raise err
    #         y = closenesses
    #         if len(X) > 1:
    #             intercept, slope, r_sq = self._2D_linear_regression(X, y)
    #             r_sq = r_sq if not np.isnan(r_sq) else 0
    #         else:
    #             intercept, slope, r_sq = 0,0,0
    #         self._features.setValByIndex(feature_name='closenessIntercept', index=lvl, new_value=intercept)
    #         self._features.setValByIndex(feature_name='closenessSlope', index=lvl, new_value=slope)
    #         self._features.setValByIndex(feature_name='closenessR2', index=lvl, new_value=r_sq)
        pass

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return ["Not Implemented"]

    # *** Optionally override public functions. ***



import json
import logging
import os
import re
import shutil
import sys
import traceback
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, IO, List, Optional, Set

# import local files
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.schemas.tables.TableSchema import TableSchema
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow

class Readme:
    def __init__(self, game_schema:GameSchema, table_schema:TableSchema):
        self._game_schema  : GameSchema  = game_schema
        self._table_schema : TableSchema = table_schema
        self._custom_src   : str   = self._getCustomSrc()
        self._dataset_meta : str   = self._getDatasetMetadata()
        self._changelog    : str   = self._getDatabaseChangelog()

    @property
    def GameName(self):
        return self._game_schema.GameName

    @property
    def CustomReadmeSource(self) -> str:
        return self._custom_src

    @property
    def DatasetMetadata(self) -> str:
        return self._dataset_meta

    @property
    def DatasetChangelog(self) -> str:
        return self._changelog

    def GenerateReadme(self, path:Path = Path("./")):
        try:
            os.makedirs(name=path, exist_ok=True)
            with open(path / "README.md", "w") as readme:
                # 1. Open files with game-specific readme data, and global db changelog.
                # 2. Use schema to write feature & column descriptions to the readme.
                readme.write("\n\n".join([
                    f"# Game: {self.GameName}",
                    self.CustomReadmeSource,
                    self.DatasetMetadata,
                    self._table_schema.AsMarkdown,
                    self._game_schema.AsMarkdown,
                    self.DatasetChangelog,
                    ""
                ]))
                # 3. Append any important data from the data changelog.
        except FileNotFoundError as err:
            Logger.Log(f"Could not open README.md for writing.", logging.ERROR)
            traceback.print_tb(err.__traceback__)
        else:
            Logger.Log(f"Wrote readme file to {path}/README.md", logging.INFO)

    def _getCustomSrc(self) -> str:
        ret_val = "No game-specific readme content prepared"

        game_schema_dir = Path(f"./games/{self.GameName}/schemas")
        try:
            with open(game_schema_dir / f"{self.GameName}_readme_src.md", "r") as readme_src:
                ret_val = readme_src.read()
        except FileNotFoundError as err:
            Logger.Log(f"Could not find {self._game_schema.GameName}_readme_src", logging.WARNING)
        finally:
            return ret_val

    def _getDatasetMetadata(self) -> str:
        """Function to generate markdown-formatted metadata for a given game.
        Gives a summary of the data licensing and suggested citation,
        then adds the markdown-formatted information from game and table schemas.

        :param game_schema: [description]
        :type game_schema: GameSchema
        :param table_schema: [description]
        :type table_schema: TableSchema
        :return: A string containing metadata for the given game.
        :rtype: str
        """
        return "\n\n".join([
            "## Field Day Open Game Data",
            "\n".join([
                "Retrieved from https://fielddaylab.wisc.edu/opengamedata",
                "These anonymous data are provided in service of future educational data mining research.",
                "They are made available under the Creative Commons CCO 1.0 Universal license.",
                "See https://creativecommons.org/publicdomain/zero/1.0/",
            ]),
            "### Suggested citation",
            "#### Field Day. (2019). Open Educational Game Play Logs - [dataset ID]. Retrieved [today's date] from https://opengamedata.fielddaylab.wisc.edu/"
        ])

    def _getDatabaseChangelog(self) -> str:
        ret_val = "No changelog prepared"

        changelog_dir = Path(f"./schemas/")
        try:
            with open(changelog_dir / "database_changelog_src.md", "r") as changelog_src:
                ret_val = changelog_src.read()
        except FileNotFoundError as err:
            Logger.Log(f"Could not find database_changelog_src", logging.WARNING)
        finally:
            return ret_val


"""
TestConfigSchema

Contains a Schema class for managing config data for testing configurations.
In particular, base testing config files always have a `"VERBOSE"` setting,
and a listing of `"ENABLED"` tests.
"""

# import standard libraries
import logging
from typing import Any, Dict, Optional

# import 3rd-party libraries

# import OGD libraries
from ogd.core.schemas.Schema import Schema

# import local files

class TestConfigSchema(Schema):
    @staticmethod
    def DEFAULT():
        return TestConfigSchema(
            name            = "DefaultTestConfig",
            verbose         = False,
            enabled_tests   = {
            }
        )

    def __init__(self, name:str, verbose:bool, enabled_tests:Dict[str, bool], other_elements:Dict[str, Any]={}):
        self._verbose       : bool            = verbose
        self._enabled_tests : Dict[str, bool] = enabled_tests
        super().__init__(name=name, other_elements=other_elements)

    @staticmethod
    def FromDict(name:str, all_elements:Dict[str, Any], logger:Optional[logging.Logger]):
        _verbose         : bool
        _enabled_tests   : Dict[str, bool]
        if "VERBOSE" in all_elements.keys():
            _verbose = TestConfigSchema._parseVerbose(all_elements["VERBOSE"], logger=logger)
        else:
            _verbose = TestConfigSchema.DEFAULT().Verbose
            _msg = f"{name} config does not have an 'VERBOSE' element; defaulting to verbose={_verbose}"
            if logger:
                logger.warn(_msg, logging.WARN)
            else:
                print(_msg)
        if "ENABLED" in all_elements.keys():
            _enabled_tests = TestConfigSchema._parseEnabledTests(all_elements["ENABLED"], logger=logger)
        else:
            _enabled_tests = TestConfigSchema.DEFAULT().EnabledTests
            _msg = f"{name} config does not have an 'ENABLED' element; defaulting to enabled={_enabled_tests}"
            if logger:
                logger.warn(_msg, logging.WARN)
            else:
                print(_msg)

        _used = {"VERBOSE", "ENABLED"}
        _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        return TestConfigSchema(name=name, verbose=_verbose, enabled_tests=_enabled_tests, other_elements=_leftovers)

    @property
    def Verbose(self) -> bool:
        return self._verbose

    @property
    def EnabledTests(self) -> Dict[str, bool]:
        return self._enabled_tests

    @property
    def AsMarkdown(self) -> str:
        ret_val : str

        ret_val = f"{self.Name}"
        return ret_val

    @staticmethod
    def _parseVerbose(verbose, logger:Optional[logging.Logger]) -> bool:
        ret_val : bool
        if isinstance(verbose, bool):
            ret_val = verbose
        elif isinstance(verbose, int):
            ret_val = bool(verbose)
        elif isinstance(verbose, str):
            ret_val = False if verbose.upper()=="FALSE" else bool(verbose)
        else:
            ret_val = bool(verbose)
            _msg = f"Config 'verbose' setting was unexpected type {type(verbose)}, defaulting to bool(verbose)={ret_val}."
            if logger:
                logger.warn(_msg, logging.WARN)
            else:
                print(_msg)
        return ret_val

    @staticmethod
    def _parseEnabledTests(enabled, logger:Optional[logging.Logger]) -> Dict[str, bool]:
        ret_val : Dict[str, bool]
        if isinstance(enabled, dict):
            ret_val = { str(key) : bool(val) for key, val in enabled.items() }
        else:
            ret_val = TestConfigSchema.DEFAULT().EnabledTests
            _msg = f"Config 'enabled tests' setting was unexpected type {type(enabled)}, defaulting to class default = {ret_val}."
            if logger:
                logger.warn(_msg, logging.WARN)
            else:
                print(_msg)
        return ret_val



__all__ = [
    "Extractor",
    "Feature",
    "PerCountFeature",
    "PerLevelFeature",
    "SessionFeature",
]

from . import Extractor
from . import Feature
from . import PerCountFeature
from . import PerLevelFeature
from . import SessionFeature


"""Initializer for Jo Wilder features"""

__all__ = [
    "QuestionAnswers",
    "SurveyItem",
    "Interaction",
    "SurveyTime",
    "Clicks",
    "Hovers",
    "SessionDuration",
    "InteractionName",
    "NotebookUses",
    "EventCount",
    "UserEnabled",
    "GameVersion",
    "UsedSaveCode",
    "GameScript",
    "SessionStart",
    "IdleState",
    "ActiveStateTime",
    "MeaningfulActions",
    "FirstInteraction",
    "LastInteraction",
    "UsedContinue",
    "InteractionWordsPerSecond",
    "InteractionTextBoxesPerSecond"
]

from . import QuestionAnswers
from . import SurveyItem
from . import Interaction
from . import SurveyTime
from . import Clicks
from . import Hovers
from . import SessionDuration
from . import InteractionName
from . import NotebookUses
from . import EventCount
from . import UserEnabled
from . import GameVersion
from . import UsedSaveCode
from . import GameScript
from . import SessionStart
from . import IdleState
from . import ActiveStateTime
from . import MeaningfulActions
from . import FirstInteraction
from . import LastInteraction
from . import UsedContinue
from . import InteractionWordsPerSecond
from . import InteractionTextBoxesPerSecond



from typing import Any, List

from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class JobsCompleted(SessionFeature):

    def __init__(self, params:GeneratorParameters, session_id:str):
        self._session_id = session_id
        super().__init__(params=params)
        self._jobs_completed = []

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["checkpoint"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        if event.EventData["status"] == "Case Closed" and event.SessionID == self._session_id:
            self._jobs_completed.append(event.EventData["mission_id"])

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._jobs_completed]

    # *** Optionally override public functions. ***



"""DataOuterface Module
"""
## import standard libraries
import abc
import logging
from typing import Any, Dict, List, Set

# import local files
from ogd.core.interfaces.Interface import Interface
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow

class DataOuterface(Interface):
    """Base class for feature and event output.

    :param Interface: _description_
    :type Interface: _type_
    :return: _description_
    :rtype: _type_
    """

    # *** ABSTRACTS ***

    @abc.abstractmethod
    def _destination(self, mode:ExportMode) -> str:
        pass

    @abc.abstractmethod
    def _removeExportMode(self, mode:ExportMode) -> str:
        pass

    @abc.abstractmethod
    def _writeRawEventsHeader(self, header:List[str]) -> None:
        pass

    @abc.abstractmethod
    def _writeProcessedEventsHeader(self, header:List[str]) -> None:
        pass

    @abc.abstractmethod
    def _writeSessionHeader(self, header:List[str]) -> None:
        pass

    @abc.abstractmethod
    def _writePlayerHeader(self, header:List[str]) -> None:
        pass

    @abc.abstractmethod
    def _writePopulationHeader(self, header:List[str]) -> None:
        pass

    @abc.abstractmethod
    def _writeRawEventLines(self, events:List[ExportRow]) -> None:
        pass

    @abc.abstractmethod
    def _writeProcessedEventLines(self, events:List[ExportRow]) -> None:
        pass

    @abc.abstractmethod
    def _writeSessionLines(self, sessions:List[ExportRow]) -> None:
        pass

    @abc.abstractmethod
    def _writePlayerLines(self, players:List[ExportRow]) -> None:
        pass

    @abc.abstractmethod
    def _writePopulationLines(self, populations:List[ExportRow]) -> None:
        pass

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_id, config:GameSourceSchema, export_modes:Set[ExportMode]):
        super().__init__(config=config)
        self._game_id : str  = game_id
        self._modes   : Set[ExportMode] = export_modes
        self._session_ct : int = 0

    def __del__(self):
        self.Close()

    @property
    def ExportModes(self) -> Set[ExportMode]:
        return self._modes

    @property
    def SessionCount(self) -> int:
        return self._session_ct
    @SessionCount.setter
    def SessionCount(self, new_val) -> None:
        self._session_ct = new_val

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def Destination(self, mode:ExportMode):
        return self._destination(mode=mode)

    def RemoveExportMode(self, mode:ExportMode):
        self._removeExportMode(mode)
        self._modes.discard(mode)
        Logger.Log(f"Removed mode {mode} from {type(self).__name__} output.", logging.INFO)

    def WriteHeader(self, header:List[str], mode:ExportMode):
        if mode in self.ExportModes:
            match (mode):
                case ExportMode.EVENTS: 
                    self._writeRawEventsHeader(header=header)
                    Logger.Log(f"Wrote event header for {self._game_id} events", depth=3)
                case ExportMode.DETECTORS:
                    self._writeProcessedEventsHeader(header=header)
                    Logger.Log(f"Wrote processed event header for {self._game_id} events", depth=3)
                case ExportMode.SESSION:
                    self._writeSessionHeader(header=header)
                    Logger.Log(f"Wrote session feature header for {self._game_id} sessions", depth=3)
                case ExportMode.PLAYER:
                    self._writePlayerHeader(header=header)
                    Logger.Log(f"Wrote player feature header for {self._game_id} players", depth=3)
                case ExportMode.POPULATION:
                    self._writePopulationHeader(header=header)
                    Logger.Log(f"Wrote population feature header for {self._game_id} populations", depth=3)
                case _:
                    Logger.Log(f"Failed to write header for unrecognized export mode {mode}!", level=logging.WARN, depth=3)
        else:
            Logger.Log(f"Skipping WriteLines in {type(self).__name__}, export mode {mode} is not enabled for this outerface", depth=3)

    def WriteLines(self, lines:List[ExportRow], mode:ExportMode) -> None:
        if mode in self.ExportModes:
            match (mode):
                case ExportMode.EVENTS:
                    self._writeRawEventLines(events=lines)
                    Logger.Log(f"Wrote {len(lines)} {self._game_id} events", depth=3)
                case ExportMode.DETECTORS:
                    self._writeProcessedEventLines(events=lines)
                    Logger.Log(f"Wrote {len(lines)} {self._game_id} processed events", depth=3)
                case ExportMode.SESSION:
                    self._writeSessionLines(sessions=lines)
                    Logger.Log(f"Wrote {len(lines)} {self._game_id} session lines", depth=3)
                case ExportMode.PLAYER:
                    self._writePlayerLines(players=lines)
                    Logger.Log(f"Wrote {len(lines)} {self._game_id} player lines", depth=3)
                case ExportMode.POPULATION:
                    self._writePopulationLines(populations=lines)
                    Logger.Log(f"Wrote {len(lines)} {self._game_id} population lines", depth=3)
                case _:
                    Logger.Log(f"Failed to write lines for unrecognized export mode {mode}!", level=logging.WARN, depth=3)
        else:
            Logger.Log(f"Skipping WriteLines in {type(self).__name__}, export mode {mode} is not enabled for this outerface", depth=3)

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData


class QuestionAnswered(Feature):
    def __init__(self, params:GeneratorParameters):
        Feature.__init__(self, params=params)
        self._answer = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.3"]
        # return ["QUESTION_ANSWER"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventData['question'] == self.CountIndex:
            self._answer = event.EventData['answered']

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._answer]

    # *** Optionally override public functions. ***



# import libraries
import json
from datetime import datetime, timedelta
from os import environ
from typing import Any, List, Optional
import numpy as np
from ogd.core.generators.Generator import GeneratorParameters
# import local files
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.games.JOWILDER import Jowilder_Enumerators as je
from ogd.games.JOWILDER.features.Interaction import clicks_track
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

# NOTE: Assumptions are: 1. All click events occured in the order like xxxx111xx222x1x3. 2. Use "text_fqid" to identify interactions. 3. The first interaction "tunic.historicalsociety.closet.intro" makes no sense so we don't need to consider it. That is, there are 190 interactions in total, but we only count 189. And we should confirm that, this tunic.historicalsociety.closet.intro doesn't occur anywhere else.

class InteractionTextBoxesPerSecond(PerCountFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self._interaction : Optional[int] = None
        self._interaction_time : List[float] = []
        self._boxes_persecond : Optional[float] = None
        self._bps_variance : Optional[float] = None
        self._first_encounter : Optional[float] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _validateEventCountIndex(self, event: Event):
        if event.EventName == "CUSTOM.1":
            return True
        _fqid = event.EventData.get("text_fqid") or event.EventData.get("cur_cmd_fqid") or "FQID NOT FOUND"
        self._interaction = je.fqid_to_enum.get(_fqid)
        if self._interaction is None:
            return self.CountIndex == clicks_track.LastInteractionIndex
        else:
            return self._interaction == self.CountIndex

        

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        # NOTE: Count all the click events
        return [f"CUSTOM.{i}" for i in range(3,12)] + ["CUSTOM.1"]
        # CUSTOM.X, X in [3,12) = ['navigate_click','notebook_click', 'map_click', 'notification_click', 'object_click', 'observation_click', 'person_click', 'cutscene_click', 'wildcard_click']

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["InteractionName"]

    def _updateFromEvent(self, event: Event) -> None:
        if not clicks_track.GameStart:
            if event.EventName == "CUSTOM.1":
                clicks_track.startGame(event)
                return
            else:
                raise(ValueError("A startgame event needed!"))
        elif event.EventName == "CUSTOM.1": 
            if clicks_track._this_click is not None and clicks_track.EventEq(event, clicks_track._this_click):
                return
            # else:
            #     raise(ValueError("Too many startgame events!"))

        if event.EventName == "CUSTOM.11" and event.EventData.get("cur_cmd_type") == 2:
            return
        if self._interaction is None:
            clicks_track.Update(event)
            clicks_track._search_state = 0
            return
        
        _time_between = clicks_track.Update(event).total_seconds()
        if clicks_track.StartNewInteraction(self._interaction):
            clicks_track._search_state = 1
            self._interaction_time.append(_time_between)
        else:
            self._interaction_time[-1] += _time_between
        
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        if len(self._interaction_time) == 0:
            return
        if feature.CountIndex != self.CountIndex:
            return
        _boxes = feature.FeatureValues[1]
        
        self._boxes_persecond = _boxes/np.mean(self._interaction_time)
        self._first_encounter = _boxes/self._interaction_time[0]
        self._bps_variance = np.var(_boxes/np.array(self._interaction_time))
        return

    def _getFeatureValues(self) -> List[Any]:
        
        return [self._boxes_persecond, self._first_encounter, self._bps_variance]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["FirstEncounter", "Variance"]




# import libraries
from datetime import timedelta
from typing import Any, List
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class UserAvgActiveTime(SessionFeature):

    def __init__(self, params:GeneratorParameters, player_id:str):
        self._player_id = player_id
        super().__init__(params=params)
        self._times : List[float] = []

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["ActiveTime"]

    def _updateFromEvent(self, event:Event) -> None:
        return

    def _updateFromFeatureData(self, feature:FeatureData):
        if feature.PlayerID == self._player_id:
            if feature.FeatureValues[0] == "No events":
                pass
            else:
                self._times.append(feature.FeatureValues[0]/timedelta(seconds=1))

    def _getFeatureValues(self) -> List[Any]:
        if len(self._times) > 0:
            return [sum(self._times) / len(self._times)]
        else:
            return [0]

    # *** Optionally override public functions. ***



import typing
from statistics import median
from typing import List, Any

from datetime import datetime as dt
from models.SequenceModel import SequenceModel

EMOTE_DESPERATE_ENUM = 2
FB_DEATH_ENUM = 18
EMOTE_EVT_ENUM = 24
MONEY_RATE_ENUM = 33
FOOD_AV_ENUM = 32

## @class SimpleDeathPredModel
# Returns the number of deaths caused by hunger, or None if no deaths due to hunger
# @param levels: Levels applicable for model

class SimpleDeathPredModel(SequenceModel):
    def __init__(self, levels: typing.List[int] = []):
        super().__init__(levels)

    def collect_deaths(self, events):
        fb_sess_cache_list = list()
        for row in reversed(events):
            if row['event_custom'] == FB_DEATH_ENUM:
                fb, sess = row['event_data_complex']['farmbit'][2], row['sess_id']
                dh_time = self.get_utc(row['client_time'])
                x = {"name": fb, "sess_id": sess, "dh_time": dh_time}
                if not len(fb_sess_cache_list):
                    fb_sess_cache_list.append(x)
                else:
                    dupes = set([el != x for el in fb_sess_cache_list])
                    unlogged = False if len(list(dupes)) > 1 else list(dupes)[0]
                    if unlogged:
                        fb_sess_cache_list.append(x)
        return fb_sess_cache_list

    def get_utc(self, data):
        dt_string = '%Y-%m-%d %H:%M:%S'
        return dt.strptime(data, dt_string).timestamp()

    def _eval(self, events, verbose=False):
        assert events
        spdh_cnt = 0
        fb_sess_cache_list = list()
        for row in reversed(events):
            if row['event_custom'] == FB_DEATH_ENUM:
                fb = row['event_data_complex']['farmbit'][2]
                sess = row['sess_id']
                x = {"name": fb, "sess_id": sess}
                if not len(fb_sess_cache_list):
                    fb_sess_cache_list.append(x)
                else:
                    dupes = set([el != x for el in fb_sess_cache_list])
                    unlogged = False if len(list(dupes)) > 1 else list(dupes)[0]
                    if unlogged:
                        fb_sess_cache_list.append(x)
            if row['event_custom'] == EMOTE_EVT_ENUM:
                if row['event_data_complex']['emote_enum'] == EMOTE_DESPERATE_ENUM:
                    fb = row['event_data_complex']['farmbit'][2]
                    sess = row['sess_id']
                    x = {"name": fb, "sess_id": sess}
                    for i, el in enumerate(fb_sess_cache_list):
                        if x == el:
                            spdh_cnt += 1
        if not len(fb_sess_cache_list):
            return None
        else:
            return spdh_cnt

## @class SimpleMoneyDeathPredModel
# Returns a dictionary of money data X seconds before deaths, or None if no logs were found in the timeframes
# @param levels: Levels applicable for model, tbd: Time before death to collect money logs
class SimpleMoneyDeathPredModel(SimpleDeathPredModel):

    def __init__(self, levels: typing.List[int] = [], tbd=60):
        self.tbd = tbd
        super().__init__(levels)

    # money
    def _eval(self, events, verbose=False):
        assert events
        money_log = list()
        fb_sess_cache_list = self.collect_deaths(events)
        for el in fb_sess_cache_list:
            for row in events:
                if row['event_custom'] == MONEY_RATE_ENUM:
                    log_time = self.get_utc(row['client_time'])
                    if (log_time < el['dh_time']) and (log_time > el['dh_time'] - self.tbd):
                        money_log.append(row['event_data_complex']['money'])
        # max, avg, min, median
        if not len(money_log):
            return {"max": None, "avg": None, "min": None,
                    "median": None}
        return {"max": max(money_log), "avg": sum(money_log) / len(money_log), "min": min(money_log),
                "median": median(money_log)}

## @class SimpleFoodDeathPredModel
# Returns a dictionary of food data X seconds before deaths, or None if no logs were found in the timeframes
# @param levels: Levels applicable for model, tbd: Time before death to collect food logs
class SimpleFoodDeathPredModel(SimpleDeathPredModel):

    def __init__(self, levels: typing.List[int] = [], tbd=60):
        self.tbd = tbd
        super().__init__(levels)

    def _eval(self, events, verbose=False):
        assert events
        food_log = list()
        fb_sess_cache_list = self.collect_deaths(events)
        for el in fb_sess_cache_list:
            for row in events:
                if row['event_custom'] == FOOD_AV_ENUM:
                    log_time = self.get_utc(row['client_time'])
                    if (log_time < el['dh_time']) and (log_time > (el['dh_time'] - self.tbd)):
                        food_log.append(row['event_data_complex']['food'])
        # max, avg, min, median
        if not len(food_log):
            return {"max": None, "avg": None, "min": None,
                    "median": None}
        return {"max": max(food_log) or 0, "avg": sum(food_log) / len(food_log), "min": min(food_log),
                "median": median(food_log)}

#     def all_hunger_death_eval(self):
#         sessions = list(self.df['sess_id'].drop_duplicates())
#         res = True
#         for sess in sessions:
#             self.hunger_death_sess_eval(sess)
#
#     def food_money_death_eval(self, event: int):
#         self.filtered_df = self.filtered_df.append(
#             self.df.query(f'(event_custom == {event})'))
#         (usr_val, cat) = (self.money, "money") if event == MONEY_RATE_ENUM else (self.food, "food")
#         for x in self.farmBits.values():
#             timeslice = self.filtered_df.query(
#                 f'(event_custom == {event}) & (utc_time_secs < {x.dh_time}) & (utc_time_secs > {x.hg_time})')
#             if len(timeslice.index) > 0:
#                 for (key, data) in zip(timeslice['utc_time_secs'], timeslice['event_data_complex']):
#                     if event == MONEY_RATE_ENUM:
#                         x.money_data[key] = data
#                     else:
#                         x.food_data[key] = data
#
#                     if data[cat] < usr_val: return False
#         return True
#
#     def hunger_death_sess_eval(self, sess: int):
#         for i, x in enumerate(self.filtered_df[(self.filtered_df['sess_id'] == sess)].itertuples()):
#             farmbit = x.event_data_complex['farmbit'][2]
#             key = f'{sess}_{farmbit}'
#             if x.event_custom == EMOTE_EVT_ENUM:
#                 self.farmBits[key] = self.FarmbitDeaths(x)
#             elif key not in self.farmBits.keys():
#                 raise Exception(
#                     f'Farmbit {farmbit} died without getting hungry. There\'s a bug, honey')
#             else:
#                 self.farmBits[key].set_dh(x)
#
#     def process_df(self):
#         date_string = '%Y-%m-%d %H:%M:%S'
#         if set(self.df_filter).issubset(set(self.df.columns)):
#             self.df = self.df.drop(columns=self.df_filter, axis=0)
#             self.df['utc_time_secs'] = self.df['client_time'].apply(
#                 lambda x: dt.strptime(x, date_string).timestamp())
#             self.process_hunger_df()
#
#     def process_hunger_df(self):
#         self.filtered_df = self.df.query(self.query)
#         self.preprocess_df()
#         temp = self.filtered_df[
#             (self.filtered_df['event_custom'] == 18) |
#             (self.filtered_df['event_custom'] == 24) &
#             ((self.filtered_df['emote_enum'] == 7) |
#              (self.filtered_df['emote_enum'] == 2))]
#         self.filtered_df = pd.DataFrame()
#         self.filtered_df = temp
#
#     def preprocess_df(self):
#         self.thresh_df = pd.DataFrame(
#             list(self.filtered_df['event_data_complex'].values))
#         farmbit = ['tile.tx', 'tile.ty', 'name', 'job_state',
#                    'job_type', 'fullness', 'energy', 'joy', 'fulfillment']
#         self.filtered_df['n_row'] = [i for i in range(len(self.filtered_df.index))]
#         self.thresh_df['n_row'] = [i for i in range(len(self.thresh_df.index))]
#
#         self.thresh_df[farmbit] = pd.DataFrame(
#             self.thresh_df.farmbit.tolist(), index=self.thresh_df.index)
#         self.filtered_df = self.filtered_df.merge(
#             self.thresh_df, on=['n_row', 'event_custom'], how='outer'
#         )
#
#     def get_farmbit_data(self, x):
#         return {
#             "fb_name": x.name,
#             "hg_dh_interval_secs": x.get_interval(),
#             "sess_id": x.sess_id,
#             "hg_data": x.hunger_data,
#             "dh_data": x.death_data}
#
#     def fb_to_csv(self, out):
#         fb_df = pd.DataFrame()
#         fb_df = fb_df.append([self.get_farmbit_data(x) for x in self.farmBits.values()], ignore_index=True)
#         print(fb_df)
#         fb_df.to_csv(out)
#
#     class FarmbitDeaths():
#         def __init__(self, hunger_data, death_data=None):
#             self.hunger_data, self.dh_data = hunger_data, np.NaN
#             self.name = self.hunger_data.event_data_complex['farmbit'][2]
#             self.food_data, self.money_data = {}, {}
#             self.hg_time, self.dh_time = self.hunger_data.utc_time_secs, -100000000
#             self.sess_id = self.hunger_data.sess_id
#
#         def set_dh(self, death_data):
#             self.dh_data = death_data
#             self.dh_time = self.dh_data.utc_time_secs
#
#         def get_interval(self):
#             if self.dh_time < 0:
#                 return -self.dh_time
#             return self.dh_time - self.hg_time
#
#         def get_food(self):
#             return sorted([[k, v['food']] for k, v in self.food_data.items()], key=lambda x: x[0])
#
#         def get_money(self):
#             return sorted([[k, v['money']] for k, v in self.money_data.items()], key=lambda x: x[0])
#
#         def print_farmbit(self):
#             dh_t_print = self.dh_data if self.dh_data == 'No death' else self.dh_data.utc_time_secs
#             s = f'Name: {self.name}\nTime:{self.get_interval()}\nSession:{self.sess_id}\n'
#             s += f'Hunger: {self.hunger_data.utc_time_secs}\nDeath: {dh_t_print}\n'
#             s += "\n".join(f'Session: {x[0]} Food: {x[1]}' for x in self.get_food()) + "\n"
#             s += "\n".join(f'Session: {x[0]} Money: {x[1]}' for x in self.get_money())
#             print(s)
#
#
# class SimpleDeathPredModel(DeathPredModel):
#     def __init__(self, levels=[], time_int=10):
#         super().__init__(levels, time_int)
#
#     def _eval(self, rows):
#         self.df = pd.DataFrame.from_records(rows)
#         self.process_df()
#         self.all_hunger_death_eval()
#         if list(self.farmBits.values())[0].get_interval() > self.time_int:
#             return False
#         return True
#
#
# class SimpleMoneyDeathPredModel(DeathPredModel):
#     def __init__(self, levels=[], money=300):
#         super().__init__(levels, money)
#
#     def Eval(self, rows):
#         return self._eval(rows)
#
#     def _eval(self, rows):
#         self.df = pd.DataFrame.from_records(rows)
#         self.process_df()
#         self.all_hunger_death_eval()
#         return self.food_money_death_eval(MONEY_RATE_ENUM)
#
#
# class SimpleFoodDeathPredModel(DeathPredModel):
#     def __init__(self, levels=[], food=300):
#         super().__init__(levels, food)
#
#     def Eval(self, rows):
#         return self._eval(rows)
#
#     def _eval(self, rows):
#         self.df = pd.DataFrame.from_records(rows)
#         self.process_df()
#         self.all_hunger_death_eval()
#         res = self.food_money_death_eval(FOOD_AV_ENUM)
#         print(res)
#         return res



# import standard libraries
import itertools
import json
from pathlib import Path
from typing import Any, Callable, Dict, Final, List, Optional

# import local files
from ogd.games import BLOOM
from ogd.games.BLOOM.detectors import * 
from ogd.games.BLOOM.features import *
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.utils import loadJSONFile
from . import features

#EXPORT_PATH : Final[str] = "games/BLOOM/DBExport.json"

## @class BloomLoader
#  Extractor subclass for extracting features from Bloomlab game data.
class BloomLoader(GeneratorLoader):
    """Class for loading Bloom generator instances."""

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the BloomlabLoader class.
    def __init__(self, player_id:str, session_id:str, game_schema: GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        """Constructor for the BloomlabLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)

    # Load Bloomlab jobs export and map job names to integer values
    # _dbexport_path = Path(BLOOM.__file__) if Path(BLOOM.__file__).is_dir() else Path(BLOOM.__file__).parent
    # with open(_dbexport_path / "DBExport.json", "r") as file:
    # export = json.load(file)

   # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule():
        return features

    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Optional[Feature]:
        ret_val : Optional[Feature]
        # First run through aggregate features
        if extractor_params._count_index == None:
            match feature_type:
                case "ActiveTime":
                    ret_val = ActiveTime.ActiveTime(params=extractor_params, idle_threshold=schema_args.get("threshold", 30))
                case "AverageActiveTime":
                    ret_val = AverageActiveTime.AverageActiveTime(params=extractor_params)
                case "BloomAlertCount":
                    ret_val = BloomAlertCount.BloomAlertCount(params=extractor_params)
                case "BuildCount":
                    ret_val = BuildCount.BuildCount(params=extractor_params)
                case "BuildingUnlockCount":
                    ret_val = BuildingUnlockCount.BuildingUnlockCount(params=extractor_params)
                case "FailCount":
                    ret_val = FailCount.FailCount(params=extractor_params)
                case "GameCompletionStatus":
                    ret_val = GameCompletionStatus.GameCompletionStatus(params=extractor_params)
                case "NumberOfSessionsPerPlayer":
                    ret_val = NumberOfSessionsPerPlayer.NumberOfSessionsPerPlayer(params=extractor_params)
                case "SucceededThroughFailure":
                    ret_val = SucceededThroughFailure.SucceededThroughFailure(params=extractor_params)
                case "CountyUnlockCount":
                    ret_val = CountyUnlockCount.CountyUnlockCount(params=extractor_params)
                case _:
                    ret_val = None
        # then run through per-count features.
        else:
            match feature_type:
                case "CountyBloomAlertCount":
                    ret_val = CountyBloomAlertCount.CountyBloomAlertCount(params=extractor_params)
                case "CountyBuildCount":
                    ret_val = CountyBuildCount.CountyBuildCount(params=extractor_params)
                case "CountyFinalPolicySettings":
                    ret_val = CountyFinalPolicySettings.CountyFinalPolicySettings(params=extractor_params)
                case "CountyLatestMoney":
                    ret_val = CountyLatestMoney.CountyLatestMoney(params=extractor_params)
                case _:
                    ret_val = None
                
        return ret_val

    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Detector:
        ret_val : Detector
        match detector_type:
            case _:
                raise NotImplementedError(f"'{detector_type}' is not a valid detector for Bloom.")
        return ret_val

    # @staticmethod
    # def GetBloomLabCount(db_export_path:Path=Path(".") / "ogd" / "games" / "BLOOM"):
    #     db_export = loadJSONFile(filename="DBExport.json", path=db_export_path)
    #     return len(db_export.get("jobs", []))

    # @staticmethod
    # def GetBloomTaskCount(db_export_path:Path=Path(".") / "ogd" / "games" / "BLOOM"):
    #     db_export = loadJSONFile(filename="DBExport.json", path=db_export_path)
    #     list_o_lists = [job.get('tasks', []) for job in db_export.get('jobs', [])]
    #     # jobs_to_task_cts = [f"{job.get('id')}: {len(job.get('tasks', []))}" for job in db_export.get('jobs', [])]
    #     # Logger.Log(f"Task counts by job:\n{jobs_to_task_cts}", logging.DEBUG)
    #     all_tasks    = list(itertools.chain.from_iterable(list_o_lists))
    #     return len(all_tasks)



import logging
from datetime import timedelta
from multiprocessing.sharedctypes import Value
from typing import Any, List

from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.utils.Logger import Logger

class PlayerSummary(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._summary = {}

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["JobsCompleted", "SessionDuration", "SessionID"]

    def _updateFromEvent(self, event:Event) -> None:
        return

    def _updateFromFeatureData(self, feature:FeatureData):
        user_id = feature.PlayerID

        if user_id not in self._summary:
            self._summary[user_id] = {
                "active_time": 0,
                "jobs_completed": [],
                "sessions": [],
                "num_sessions": 0
            }

        if feature.ExportMode == ExtractionMode.PLAYER:
            if feature.FeatureType == "JobsCompleted":
                self._summary[user_id]["jobs_completed"] = feature.FeatureValues[0]
            elif feature.FeatureType == "SessionDuration":
                if isinstance(feature.FeatureValues[0], timedelta):
                    self._summary[user_id]["active_time"] += feature.FeatureValues[0].total_seconds()
                elif isinstance(feature.FeatureValues[0], str) and feature.FeatureValues[0] == "No events":
                    pass
                else:
                    raise ValueError(f"PlayerSummary got {feature.Name} feature with value {feature.FeatureValues[0]} of non-timedelta type {type(feature.FeatureValues[0])} in the {feature.FeatureNames[0]} column!")
        elif feature.ExportMode == ExtractionMode.SESSION:
            if feature.FeatureType == "SessionID":
                self._summary[user_id]["sessions"].append(feature.FeatureValues[0])

    def _getFeatureValues(self) -> List[Any]:
        for summary in self._summary.values():
            summary['num_sessions'] = len(set(summary['sessions']))
        return [self._summary]

    # *** Optionally override public functions. ***

    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        """List of ExtractionMode supported by the Feature.

        Overridden from base Feature version.
        A PlayerSummary is only used at player and population levels; not concerned with session-level.
        :return: _description_
        :rtype: List[ExtractionMode]
        """
        return [ExtractionMode.POPULATION, ExtractionMode.PLAYER]



"""Initializer for Shipwrecks features"""

__all__ = [
    "ActiveJobs",
    "EventList",
    "EvidenceBoardCompleteCount",
    "JobsAttempted",
    "JobsCompleted",
    "PlayerSummary",
    "PopulationSummary",
    "MissionDiveTime",
    "MissionSonarTimeToComplete",
    "SessionDuration",
    "SessionID",
    "TopJobCompletionDestinations",
    "TopJobSwitchDestinations",
    "TotalDiveTime"
]

from . import ActiveJobs
from . import EventList
from . import EvidenceBoardCompleteCount
from . import JobsAttempted
from . import JobsCompleted
from . import PlayerSummary
from . import PopulationSummary
from . import MissionDiveTime
from . import MissionSonarTimeToComplete
from . import SessionDuration
from . import SessionID
from . import TopJobCompletionDestinations
from . import TopJobSwitchDestinations
from . import TotalDiveTime



## import standard libraries
from typing import Any, Callable, Dict, List, Optional
## import local files
import ogd.games.CRYSTAL.features
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.legacy.LegacyLoader import LegacyLoader
from ogd.games.CRYSTAL.features.CrystalExtractor import CrystalExtractor
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils.Logger import Logger

class CrystalLoader(LegacyLoader):

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule():
        return ogd.games.CRYSTAL.features

    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Optional[Feature]:
        return CrystalExtractor(params=extractor_params, game_schema=self._game_schema, session_id=self._session_id)

    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Optional[Detector]:
        Logger.Log(f"'{detector_type}' is not a valid detector for Crystal.")

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the CrystalLoader class.
    def __init__(self, player_id:str, session_id:str, game_schema:GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        """Constructor for the CrystalLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
from datetime import timedelta
from typing import Any, List
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event, EventSource
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class TotalDiveTime(Feature):
    
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self.prev_time = timedelta(0)
        self.total_time = timedelta(0)
        self.idle_time = timedelta(0)
        self.on = False

    def Subfeatures(self) -> List[str]: 
        return ["Seconds", "Active", "Active-Seconds", "Idle", "Idle-Seconds"]
    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["begin_dive", "scene_changed", "end_dive"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventSource == EventSource.GAME:
            if self.on:
                self.total_time += (event.Timestamp - self.prev_time)
                if (event.Timestamp - self.prev_time) > timedelta(minutes=1):
                    self.idle_time += (event.Timestamp - self.prev_time)
                if event.EventName == 'end_dive':
                    self.on = False
            else:
                if event.EventName == "begin_dive":
                    self.on = True
            self.prev_time = event.Timestamp
       
        # if event.SessionID != self._session_id:
        #     self._session_id = event.SessionID

        #     if self._dive_start_time:
        #         self._time += (self._prev_timestamp - self._dive_start_time).total_seconds()
        #         self._dive_start_time = event.Timestamp

        # if event.EventName == "begin_dive":
        #     self._dive_start_time = event.Timestamp
        # elif event.EventName == "scene_changed":
        #     if self._dive_start_time is not None:
        #         self._time += (event.Timestamp - self._dive_start_time).total_seconds()
        #         self._dive_start_time = None

        # self._prev_timestamp = event.Timestamp

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if self.total_time is not None:
            return [self.total_time, self.total_time.total_seconds(), (self.total_time - self.idle_time), (self.total_time - self.idle_time).total_seconds(), self.idle_time, self.idle_time.total_seconds()]
        else:
            return ["No events"]

    # *** Optionally override public functions. ***



"""Initializer for Bloom features"""

__all__ = [
    "PerCountyFeature",
    "ActiveTime",
    "AverageActiveTime",
    "BloomAlertCount",
    "BuildCount",
    "BuildingUnlockCount",
    "CountyBloomAlertCount",
    "CountyBuildCount",
    "CountyFinalPolicySettings",
    "CountyLatestMoney",
    "CountyUnlockCount",
    "FailCount",
    "GameCompletionStatus",
    "NumberOfSessionsPerPlayer",
    "SucceededThroughFailure"

]

from . import PerCountyFeature
from . import ActiveTime
from . import AverageActiveTime
from . import BloomAlertCount
from . import BuildCount
from . import BuildingUnlockCount
from . import CountyBloomAlertCount
from . import CountyBuildCount
from . import CountyFinalPolicySettings
from . import CountyLatestMoney
from . import CountyUnlockCount
from . import FailCount
from . import GameCompletionStatus
from . import NumberOfSessionsPerPlayer
from . import SucceededThroughFailure



# import libraries
from collections import defaultdict
from statistics import stdev
from typing import Any, List
# import locals
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData


class JobsAttempted(Feature):

    def __init__(self, params:GeneratorParameters, mission_map:dict):
        self._mission_map = mission_map
        super().__init__(params=params)
        self._session_id = None
        self._start_map = defaultdict(list)
        self._complete_map = defaultdict(list)

        # Subfeatures
        if self.CountIndex is not None:
            self._mission_id = self.CountIndex
            self._mission_name = list(mission_map.keys())[self.CountIndex]
        else:
            raise ValueError("JobsAttempted got a count index of None!")
        self._num_starts = 0
        self._num_completes = 0
        self._percent_complete = 0
        self._avg_time_complete = 0
        self._std_dev_complete = 0

        # Time
        self._times = []
        self._mission_start_time = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["checkpoint"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        session_id = event.SessionID
        checkpoint = event.EventData["status"]
        mission_name = event.EventData["mission_id"]
        mission_id = self._mission_map[mission_name]

        if checkpoint == "Begin Mission":
            if mission_name == "Level4" and session_id not in self._complete_map[2]:
                return
            elif mission_id == self._mission_id and session_id not in self._start_map[mission_id]:
                self._num_starts += 1
                self._session_id = session_id
                self._mission_start_time = event.Timestamp
                self._start_map[mission_id].append(session_id)
                self._session_id = session_id

        elif checkpoint == "Case Closed":
            self._complete_map[mission_id].append(session_id)

            if mission_id == self._mission_id and session_id == self._session_id:
                self._num_completes += 1
                self._complete_map[mission_id].append(session_id)

                if self._mission_start_time:
                    self._times.append((event.Timestamp - self._mission_start_time).total_seconds())
                    self._mission_start_time = None

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if self._num_starts > 0:
            self._percent_complete = (self._num_completes / self._num_starts) * 100
        
        if len(self._times) > 0:
            self._avg_time_complete = sum(self._times) / len(self._times)

        if len(self._times) > 1:
            self._std_dev_complete = stdev(self._times)

        return [self._mission_id, self._mission_name, self._num_starts, self._num_completes, self._percent_complete, self._avg_time_complete, self._std_dev_complete]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["job-name", "num-starts", "num-completes", "percent-complete", "avg-time-complete", "std-dev-complete"]



# import libraries
from typing import Any, Dict, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class NumberOfSessionsPerPlayer(Feature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.session_count: Dict[str, int] = {}

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["session_start"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        player_id = event.user_id
        if player_id not in self.session_count:
            self.session_count[player_id] = 1
        else:
            self.session_count[player_id] += 1

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        total_sessions = sum(self.session_count.values())
        return [total_sessions]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



# import libraries
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class OverallSliderAverageRange(SessionFeature):
    def __init__(self, params:GeneratorParameters):
        SessionFeature.__init__(self, params=params)
        self._ranges = []

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.1"]
        # return ["SLIDER_MOVE_RELEASE"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        self._ranges.append(event.EventData["max_val"] - event.EventData["min_val"])

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if len(self._ranges) > 0:
            return [sum(self._ranges) / len(self._ranges)]
        else:
            return [None]

    # *** Optionally override public functions. ***





# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from datetime import datetime, timedelta
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.extractors.SessionFeature import SessionFeature

class EggRecoverTime(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._session_id = None
        self._stolen_timestamp : Optional[datetime] = None
        self._time = 0
        self._skua_id = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["egg_lost", "egg_recovered"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "egg_lost":
            self._skua_id = event.event_data.get("object_id")
            self._stolen_timestamp = event.Timestamp
            # Logger.Log(f"lost time is {self._argument_start_time}")
        elif event.EventName == "egg_recovered":
            if self._stolen_timestamp is not None:
                self._time += (event.Timestamp - self._stolen_timestamp).total_seconds()
                self._stolen_timestamp = None
            else: 
                Logger.Log("Got an 'egg_recovered' with no preceding 'egg_lost' event!", logging.WARN)
    
    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [timedelta(seconds=self._time)]

    


import json
import logging
import os
from datetime import datetime
from google.cloud import bigquery
from typing import Dict, List, Tuple, Optional
# import locals
from coding.Code import Code
from coding.Coder import Coder
from ogd.core.interfaces.CodingInterface import CodingInterface
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.schemas.configs.ConfigSchema import ConfigSchema
from ogd.core.utils.Logger import Logger

# TODO: see about merging this back into BigQueryInterface for a unified interface.

class BigQueryCodingInterface(CodingInterface):

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_id:str, config:ConfigSchema):
        super().__init__()
        self._game_id : str = game_id
        self._settings = config
        self.Open()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _open(self, force_reopen: bool = False) -> bool:
        if force_reopen:
            self.Close()
            self.Open(force_reopen=False)
        if not self._is_open:
            if "GITHUB_ACTIONS" in os.environ:
                self._client = bigquery.Client()
            else:
                credential_path : str
                if "GAME_SOURCE_MAP" in self._settings:
                    credential_path = self._settings["GAME_SOURCE_MAP"][self._game_id]["credential"]
                else:
                    credential_path = default_settings["GAME_SOURCE_MAP"][self._game_id]["credential"]
                os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = credential_path
                self._client = bigquery.Client()
            if self._client != None:
                self._is_open = True
                Logger.Log("Connected to BigQuery database.", logging.DEBUG)
                return True
            else:
                Logger.Log("Could not connect to BigQuery Database.", logging.WARN)
                return False
        else:
            return True

    def _close(self) -> bool:
        self._client.close()
        self._is_open = False
        Logger.Log("Closed connection to BigQuery.", logging.DEBUG)
        return True

    def _allCoders(self) -> Optional[List[Coder]]:
        query = f"""
            SELECT DISTINCT coder_id, name
            FROM `{self._dbPath()}.coders`
        """
        data = self._client.query(query)
        coders = [Coder(name=str(row['name']), id=str(row['coder_id'])) for row in data]
        return coders or []

    def _createCoder(self, coder_name:str) -> bool:
        # TODO: figure out how to make metadata available for insert.
        query = f"""
            INSERT {self._dbPath()}(coder_id, name, metadata)
            VALUES (GENERATE_UUID(), @name, NULL)
        """
        cfg = bigquery.QueryJobConfig(
            query_parameters= [
                bigquery.ScalarQueryParameter(name="name", type_="STRING", value=coder_name),
            ]
        )
        try:
            self._client.query(query=query, job_config=cfg)
        except Exception as err:
            Logger.Log(f"Error while creating a new Coder in database: {err}", level=logging.ERROR, depth=2)
            return False
        else:
            return True

    def _getCodeWordsByGame(self, game_id:str) -> Optional[List[str]]:
        query = f"""
            SELECT DISTINCT code
            FROM `{self._dbPath(game_id=game_id)}.codes`
        """
        data = self._client.query(query)
        codes = [str(row['code']) for row in data]
        return codes or []

    def _getCodeWordsByCoder(self, coder_id:str) -> Optional[List[str]]:
        query = f"""
            SELECT DISTINCT code
            FROM `{self._dbPath()}.codes`
            WHERE coder_id=@coder_id
        """
        cfg = bigquery.QueryJobConfig(
            query_parameters= [
                bigquery.ScalarQueryParameter(name="coder_id", type_="STRING", value=coder_id),
            ]
        )
        data = self._client.query(query=query, job_config=cfg)
        codes = [str(row['code']) for row in data]
        return codes or []

    def _getCodeWordsBySession(self, session_id:str) -> Optional[List[str]]:
        query = f"""
            SELECT DISTINCT code
            FROM `{self._dbPath()}.codes`
            CROSS JOIN UNNEST(event_params) AS param_session
            WHERE param_session.key = 'ga_session_id' AND param_session.value.int_value=@session_id)
        """
        i_session_id : int = int(session_id)
        cfg = bigquery.QueryJobConfig(
            query_parameters= [
                bigquery.ScalarQueryParameter(name="session_id", type_="INTEGER", value=i_session_id),
            ]
        )
        data = self._client.query(query=query, job_config=cfg)
        codes = [str(row['code']) for row in data]
        return codes or []

    def _getCodesByGame(self, game_id:str) -> Optional[List[Code]]:
        query = f"""
            SELECT *
            FROM `{self._dbPath(game_id=game_id)}.codes`,
        """
        try:
            data = self._client.query(query)
            codes = [BigQueryCodingInterface._codeFromRow(row=row) for row in data]
        except Exception as err:
            Logger.Log(f"Error while retrieving {game_id} codes from database: {err}", level=logging.ERROR, depth=2)
            return []
        else:
            return codes

    def _getCodesByCoder(self, coder_id:str) -> Optional[List[Code]]:
        query = f"""
            SELECT *
            FROM `{self._dbPath()}.codes`
            WHERE `coder_id`=@coder_id
        """
        cfg = bigquery.QueryJobConfig(
            query_parameters= [
                bigquery.ScalarQueryParameter(name="coder_id", type_="STRING", value=coder_id),
            ]
        )
        try:
            data = self._client.query(query)
            codes = [BigQueryCodingInterface._codeFromRow(row=row) for row in data]
        except Exception as err:
            Logger.Log(f"Error while retrieving {coder_id} codes from database: {err}", level=logging.ERROR, depth=2)
            return []
        else:
            return codes

    def _getCodesBySession(self, session_id:str) -> Optional[List[Code]]:
        query = f"""
            SELECT *
            FROM `{self._dbPath()}.codes`
            CROSS JOIN UNNEST(events) AS param_session
            WHERE param_session.key = 'session_id' AND param_session.value.int_value=@session_id
        """
        cfg = bigquery.QueryJobConfig(
            query_parameters= [
                bigquery.ScalarQueryParameter(name="session_id", type_="STRING", value=session_id),
            ]
        )
        try:
            data = self._client.query(query)
            codes = [BigQueryCodingInterface._codeFromRow(row=row) for row in data]
        except Exception as err:
            Logger.Log(f"Error while retrieving {session_id} codes from database: {err}", level=logging.ERROR, depth=2)
            return []
        else:
            return codes

    def _createCode(self, code:str, coder_id:str, events:List[Code.EventID], notes:Optional[str]=None):
        query = f"""
            INSERT {self._dbPath()}(code_id, code, coder_id, notes, events)
            VALUES (GENERATE_UUID(), @code, @coder_id, @notes, @events)
        """
        evt_params = [
            bigquery.StructQueryParameter.positional(
                bigquery.ScalarQueryParameter(name="session_id", type_="STRING", value=event.SessionID),
                bigquery.ScalarQueryParameter(name="index", type_="INTEGER", value=event.Index)
            )
            for event in events
        ]
        cfg = bigquery.QueryJobConfig(
            query_parameters= [
                bigquery.ScalarQueryParameter(name="code", type_="STRING", value=code),
                bigquery.ScalarQueryParameter(name="coder_id", type_="STRING", value=coder_id),
                bigquery.ScalarQueryParameter(name="notes", type_="STRING", value=notes),
                bigquery.ArrayQueryParameter(
                    name="events",
                    array_type="STRUCT",
                    values=evt_params
                ),
            ]
        )
        try:
            self._client.query(query=query, job_config=cfg)
        except Exception as err:
            Logger.Log(f"Error while creating a new Coder in database: {err}", level=logging.ERROR, depth=2)
            return False
        else:
            return True

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def IsOpen(self) -> bool:
        """Overridden version of IsOpen function, checks that BigQueryInterface client has been initialized.

        :return: True if the interface is open, else False
        :rtype: bool
        """
        return True if (super().IsOpen() and self._client is not None) else False

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    @staticmethod
    def _codeFromRow(row) -> Code:
        _code_word = str(row['code'])
        _code_id   = str(row['code_id'])
        _coder_id  = str(row['coder_id'])
        _name      = str(row['coder_name'])
        _events    = [] # TODO: figure out how to parse these out.
        _notes     = str(row['notes'])
        return Code(code_word=_code_word, id=_code_id,
                    coder=Coder(name=_name, id=_coder_id),
                    events=_events, notes=_notes
        )

    # *** PRIVATE METHODS ***

    def _dbPath(self, game_id:Optional[str]=None) -> str:
        _game_id = game_id or self._game_id
        if "BIGQUERY_CONFIG" in self._settings:
            project_name = self._settings["BIGQUERY_CONFIG"][_game_id]["PROJECT_ID"]
        else:
            project_name = default_settings["BIGQUERY_CONFIG"][_game_id]["PROJECT_ID"]
        return f"{project_name}.coding"



# import standard libraries
from typing import Any, Dict, List
# import local files
from ogd.core.schemas.games.FeatureSchema import FeatureSchema

class AggregateSchema(FeatureSchema):
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        super().__init__(name=name, all_elements=all_elements)

    @property
    def AsMarkdown(self) -> str:
        ret_val   : str

        ret_val = f"**{self.Name}** : *{self.ReturnType}*, *Aggregate feature* {' (disabled)' if not len(self.Enabled) > 0 else ''}  \n{self.Description}  \n"
        if len(self.Subfeatures) > 0:
            ret_val += "*Sub-features*:  \n\n" + "\n".join([subfeature.AsMarkdown for subfeature in self.Subfeatures.values()])
        if len(self.NonStandardElements) > 0:
            ret_val += "*Other elements*:  \n\n" + "\n".join([f"{elem_name} : {elem}" for elem_name,elem in self.NonStandardElements.items()])
        return ret_val



# import standard libraries
import logging
from typing import Any, Dict, Optional, Type
# import local files
from ogd.core.schemas.Schema import Schema
from ogd.core.schemas.configs.data_sources.DataSourceSchema import DataSourceSchema
from ogd.core.utils.Logger import Logger

class SSHSchema(Schema):
    def __init__(self, name:str, all_elements:Dict[str, Any], fallbacks:Dict[str, Any]={}):
        self._host : Optional[str]
        self._user : Optional[str]
        self._pass : Optional[str]
        self._port : int

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For {name} base config, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        if "SSH_HOST" in all_elements.keys():
            self._host = SSHSchema._parseHost(all_elements["SSH_HOST"])
        elif "SSH_HOST" in fallbacks.keys():
            self._host = SSHSchema._parseHost(fallbacks["SSH_HOST"])
        else:
            self._host = None
            Logger.Log(f"{name} config does not have a 'SSH_HOST' element; defaulting to ssh_host={self._host}", logging.WARN)
        if "SSH_USER" in all_elements.keys():
            self._user = SSHSchema._parseUser(all_elements["SSH_USER"])
        elif "SSH_USER" in fallbacks.keys():
            self._user = SSHSchema._parseUser(fallbacks["SSH_USER"])
        else:
            self._user = None
            Logger.Log(f"{name} config does not have a 'SSH_USER' element; defaulting to ssh_user={self._user}", logging.WARN)
        if "SSH_PW" in all_elements.keys():
            self._pass = SSHSchema._parsePass(all_elements["SSH_PW"])
        elif "SSH_PASS" in all_elements.keys():
            self._pass = SSHSchema._parsePass(all_elements["SSH_PASS"])
        elif "SSH_PW" in fallbacks.keys():
            self._pass = SSHSchema._parsePass(fallbacks["SSH_PW"])
        elif "SSH_PASS" in fallbacks.keys():
            self._pass = SSHSchema._parsePass(fallbacks["SSH_PASS"])
        else:
            self._pass = None
            Logger.Log(f"{name} config does not have a 'SSH_PASS' or 'SSH_PW' element; defaulting to ssh_pass={self._pass}", logging.WARN)
        if "SSH_PORT" in all_elements.keys():
            self._port = SSHSchema._parsePort(all_elements["SSH_PORT"])
        elif "SSH_PORT" in fallbacks.keys():
            self._port = SSHSchema._parsePort(fallbacks["SSH_PORT"])
        else:
            self._port = 22
            Logger.Log(f"{name} config does not have a 'SSH_PORT' element; defaulting to ssh_port={self._port}", logging.WARN)

        _used = {"SSH_HOST", "SSH_USER", "SSH_PW", "SSH_PASS", "SSH_PORT"}
        _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def Host(self) -> Optional[str]:
        return self._host

    @property
    def User(self) -> Optional[str]:
        return self._user

    @property
    def Pass(self) -> Optional[str]:
        return self._pass

    @property
    def Port(self) -> int:
        return self._port

    @property
    def AsMarkdown(self) -> str:
        ret_val : str

        ret_val = f"{self.Name} : `{self.AsConnectionInfo}`"
        return ret_val

    @property
    def AsConnectionInfo(self) -> str:
        ret_val : str

        ret_val = f"{self.User}@{self.Host}:{self.Port}"
        return ret_val

    @staticmethod
    def _parseHost(host) -> Optional[str]:
        ret_val : Optional[str]
        if isinstance(host, str):
            ret_val = host
        else:
            ret_val = str(host)
            Logger.Log(f"SSH config for host was unexpected type {type(host)}, defaulting to str(host)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseUser(user) -> Optional[str]:
        ret_val : Optional[str]
        if isinstance(user, str):
            ret_val = user
        else:
            ret_val = str(user)
            Logger.Log(f"SSH config for user was unexpected type {type(user)}, defaulting to str(user)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parsePass(pw) -> Optional[str]:
        ret_val : Optional[str]
        if isinstance(pw, str):
            ret_val = pw
        else:
            ret_val = str(pw)
            Logger.Log(f"SSH config for password was unexpected type {type(pw)}, defaulting to str(pw)=***.", logging.WARN)
        return ret_val

    @staticmethod
    def _parsePort(port) -> int:
        ret_val : int
        if isinstance(port, int):
            ret_val = port
        elif isinstance(port, str):
            ret_val = int(port)
        else:
            ret_val = int(port)
            Logger.Log(f"SSH config for port was unexpected type {type(port)}, defaulting to int(port)={ret_val}.", logging.WARN)
        return ret_val

class MySQLSchema(DataSourceSchema):
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._db_host  : str
        self._db_port  : int
        self._db_user  : str
        self._db_pass  : Optional[str]
        self._ssh_cfg  : SSHSchema

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For {name} MySQL Data Source config, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        # Parse DB info
        if "DB_HOST" in all_elements.keys():
            self._db_host = MySQLSchema._parseDBHost(all_elements["DB_HOST"])
        else:
            self._db_host = "UNKNOWN"
            Logger.Log(f"{name} config does not have a 'DB_HOST' element; defaulting to db_host={self._db_host}", logging.WARN)
        if "DB_PORT" in all_elements.keys():
            self._db_port = MySQLSchema._parseDBPort(all_elements["DB_PORT"])
        else:
            self._db_port = 3306
            Logger.Log(f"{name} config does not have a 'DB_PORT' element; defaulting to db_port={self._db_port}", logging.WARN)
        if "DB_USER" in all_elements.keys():
            self._db_user = MySQLSchema._parseDBUser(all_elements["DB_USER"])
        else:
            self._db_user = "UNKOWN"
            Logger.Log(f"{name} config does not have a 'DB_USER' element; defaulting to db_user={self._db_user}", logging.WARN)
        if "DB_PW" in all_elements.keys():
            self._db_pass = MySQLSchema._parseDBPass(all_elements["DB_PW"])
        elif "DB_PASS" in all_elements.keys():
            self._db_pass = MySQLSchema._parseDBPass(all_elements["DB_PASS"])
        else:
            self._db_pass = None
            Logger.Log(f"{name} config does not have a 'DB_PW' element; defaulting to db_pass=None", logging.WARN)
        # Parse SSH info, if it exists. Don't notify, if it doesn't exist.
        _ssh_keys = {"SSH_HOST", "SSH_PORT", "SSH_USER", "SSH_PW", "SSH_PASS"}
        self._ssh_cfg = SSHSchema(name=f"{name}-SSH", all_elements={ key : all_elements.get(key) for key in _ssh_keys.intersection(all_elements.keys()) })

        _used = {"DB_HOST", "DB_PORT", "DB_USER", "DB_PW", "DB_PASS", "SSH_HOST", "SSH_PORT", "SSH_USER", "SSH_PW", "SSH_PASS"}
        _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def DBHost(self) -> str:
        return self._db_host

    @property
    def DBPort(self) -> int:
        return self._db_port

    @property
    def DBUser(self) -> str:
        return self._db_user

    @property
    def DBPass(self) -> Optional[str]:
        return self._db_pass

    @property
    def SSHConfig(self) -> SSHSchema:
        return self._ssh_cfg

    @property
    def SSH(self) -> SSHSchema:
        """Shortened alias for SSHConfig, convenient when using sub-elements of the SSHConfig.

        :return: The schema describing the configuration for an SSH connection to a data source.
        :rtype: SSHSchema
        """
        return self._ssh_cfg

    @property
    def HasSSH(self) -> bool:
        """Property indicating if this MySQL source has a valid SSH configuration attached to it.

        :return: True if there is a valid SSH configuration, otherwise false.
        :rtype: bool
        """
        return (self.SSH.Host is not None and self.SSH.User is not None and self.SSH.Pass is not None)

    @property
    def AsMarkdown(self) -> str:
        ret_val : str

        ssh_part = f"{self.SSH.AsConnectionInfo} -> " if self.HasSSH else ""
        ret_val  = f"{self.Name} : `{ssh_part}{self.AsConnectionInfo}` ({self.Type})"
        return ret_val

    @property
    def AsConnectionInfo(self) -> str:
        ret_val : str = f"{self.DBUser}@{self.DBHost}:{self.DBPort}"
        return ret_val

    @staticmethod
    def _parseDBHost(db_host) -> str:
        ret_val : str
        if isinstance(db_host, str):
            ret_val = db_host
        else:
            ret_val = str(db_host)
            Logger.Log(f"MySQL Data Source DB host was unexpected type {type(db_host)}, defaulting to str(db_host)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseDBPort(db_port) -> int:
        ret_val : int
        if isinstance(db_port, int):
            ret_val = db_port
        else:
            ret_val = int(db_port)
            Logger.Log(f"MySQL Data Source DB port was unexpected type {type(db_port)}, defaulting to int(db_port)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseDBUser(db_user) -> str:
        ret_val : str
        if isinstance(db_user, str):
            ret_val = db_user
        else:
            ret_val = str(db_user)
            Logger.Log(f"MySQL Data Source DB username was unexpected type {type(db_user)}, defaulting to str(db_user)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseDBPass(db_pass) -> str:
        ret_val : str
        if isinstance(db_pass, str):
            ret_val = db_pass
        else:
            ret_val = str(db_pass)
            Logger.Log(f"MySQL Data Source DB password was unexpected type, defaulting to str(db_pass)=***.", logging.WARN)
        return ret_val




# import libraries
import json
from typing import Any, List, Optional
# import local files
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature




class SnippetReceivedCount(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        choice_type_names = ["BAD", "GOOD", "GREAT"]
        self._quality_counts = {name: 0 for name in choice_type_names}
        self._snippet_receive_count : int = 0
        

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["snippet_received"] # >>> fill in names of events this Feature should use for extraction. <<<

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of first-order features this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        self._snippet_receive_count += 1
        _quality = event.EventData.get("snippet_quality", "QUALITY NOT FOUND").upper()
        if _quality is not None and _quality in self._quality_counts.keys():
            self._quality_counts[_quality] += 1
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        # >>> use data in the FeatureData object to update state variables as needed. <<<
        # Note: This function runs on data from each Feature whose name matches one of the strings returned by _featureFilter().
        #       The number of instances of each Feature may vary, depending on the configuration and the unit of analysis at which this CustomFeature is run.
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        return [self._snippet_receive_count, self._quality_counts["BAD"],
        self._quality_counts["GOOD"], self._quality_counts["GREAT"]]


    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Bad", "Good", "Great"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.POPULATION, ExtractionMode.PLAYER, ExtractionMode.SESSION] # >>> delete any modes you don't want run for your Feature. <<<
    
    # @staticmethod
    # def MinVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
    #     return super().MinVersion()

    # @staticmethod
    # def MaxVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
    #     return super().MaxVersion()



# import libraries
import json
from collections import defaultdict
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData


class ActiveJobs(Feature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._current_session_id = None
        self._last_started_id = None
        self._active_jobs = defaultdict(list)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["checkpoint"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        session_id = event.SessionID
        mission_name = event.EventData["mission_id"]

        if (self._current_session_id is not None) and (self._current_session_id != session_id) and (self._current_session_id not in self._active_jobs[self._last_started_id]):
            # if we found a new user, then previous user must have left off on whatever their active job was.
            # so, add the user to the list for that job
            self._active_jobs[self._last_started_id].append(self._current_session_id)
        self._current_session_id = session_id # in either case, set latest user as "current"
        self._last_started_id = mission_name # In either case, set latest job name as "current".

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        ret_val = self._active_jobs

        if (self._last_started_id is not None) and (self._current_session_id not in self._active_jobs[self._last_started_id]):
            ret_val[self._last_started_id].append(self._current_session_id) # whatever last event was, assume player left off there.

        return [json.dumps(ret_val)]

    # *** Optionally override public functions. ***



"""Module for generating certain objects for internal use in the OGD CLI"""
# import standard libraries
import logging
from calendar import monthrange
from datetime import datetime
from typing import Optional, Set

# import 3rd-party libraries

# import local files
from ogd.core.interfaces.EventInterface import EventInterface
from ogd.core.interfaces.MySQLInterface import MySQLInterface
from ogd.core.interfaces.BigQueryInterface import BigQueryInterface
from ogd.core.interfaces.BQFirebaseInterface import BQFirebaseInterface
from ogd.core.requests.Request import ExporterRange
from ogd.core.schemas.configs.ConfigSchema import ConfigSchema
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.utils.Logger import Logger

class OGDGenerators:
    """Utility class to collect functions for generating objects used to execute certain commands.
    
    Essentially, just a collection of random stuff that we didn't want cluttering other files.
    """

    @staticmethod
    def GenDBInterface(config:ConfigSchema, game:str) -> EventInterface:
        """Create a data interface based on a config and desired game.

        :param config: The current OGD configuration
        :type config: ConfigSchema
        :param game: The ID of the game whose data should be retrieved from the interface
        :type game: str
        :raises Exception: If the configuration for the given game does not give a valid type of database for the source.
        :raises ValueError: If the given game does not exist in the GameSourceMap of the given configuration.
        :return: A data interface for the configured type of database.
        :rtype: EventInterface

        .. todo:: Accept a GameSourceSchema instead of a full ConfigSchema
        .. todo:: Use the "upper" of the source type, instead of checking for capitalized and non-capitalized versions of names.
        """
        ret_val : EventInterface
        _game_cfg = config.GameSourceMap.get(game)
        if _game_cfg is not None and _game_cfg.Source is not None:
            match (_game_cfg.Source.Type):
                case "Firebase" | "FIREBASE":
                    ret_val = BQFirebaseInterface(game_id=game, config=_game_cfg, fail_fast=config.FailFast)
                case "BigQuery" | "BIGQUERY":
                    ret_val = BigQueryInterface(game_id=game, config=_game_cfg, fail_fast=config.FailFast)
                case "MySQL" | "MYSQL":
                    ret_val = MySQLInterface(game_id=game, config=_game_cfg, fail_fast=config.FailFast)
                case _:
                    raise Exception(f"{_game_cfg.Source.Type} is not a valid EventInterface type!")
            return ret_val
        else:
            raise ValueError(f"Config for {game} was invalid or not found!")

    @staticmethod
    def GenModes(with_events:bool, with_features:bool, no_session_file:bool, no_player_file:bool, no_pop_file:bool) -> Set[ExportMode]:
        """Convert a series of booleans for each type of export mode into a set of ExportMode enum values.

        :param with_events: Whether to include `EVENTS` and `DETECTORS` in the set
        :type with_events: bool
        :param with_features: Whether to include any feature types in the set
        :type with_features: bool
        :param no_session_file: Whether to include `SESSION` in the set
        :type no_session_file: bool
        :param no_player_file: Whether to include `PLAYER` in the set
        :type no_player_file: bool
        :param no_pop_file: Whether to include `POPULATION` in the set
        :type no_pop_file: bool
        :return: A set of ExportModes, based on the arguments
        :rtype: Set[ExportMode]
        """
        ret_val = set()

        if with_events:
            ret_val.add(ExportMode.EVENTS)
            ret_val.add(ExportMode.DETECTORS)
        if with_features:
            if not no_session_file:
                ret_val.add(ExportMode.SESSION)
            if not no_player_file:
                ret_val.add(ExportMode.PLAYER)
            if not no_pop_file:
                ret_val.add(ExportMode.POPULATION)

        return ret_val

    # retrieve/calculate date range.
    @staticmethod
    def GenDateRange(game:str, interface:EventInterface, monthly:bool, start_date:str, end_date:Optional[str]) -> ExporterRange:
        """Use a pair of date strings to create an `ExporterRange` for use with an interface.

        Also allows the range to be specified as "monthly,"
        i.e. to treat the "start date" as a specification of a full month for the range.
        Note that `ExporterRange` objects carry data about the sessions contained within the range,
        so an interface is required in order to create the session list.

        :param game: The specific game for which a date range is generated
        :type game: str
        :param interface: An interface to use for generation of the `ExporterRange`.
        :type interface: EventInterface
        :param monthly: Whether the range should cover a full month, or use the exact given start and end.
        :type monthly: bool
        :param start_date: A string representing the first day of the range in MM/DD/YYYY format, or the month to use for the range in MM/YYYY format.
        :type start_date: str
        :param end_date: A string representing the last day of the range in MM/DD/YYYY format, or None (if using a full month range)
        :type end_date: Optional[str]
        :raises ValueError: If using full month range, and `start_date` does not have a correct format.
        :return: An `ExporterRange` object representing the given range, as well as the sessions available for that range via the given interface.
        :rtype: ExporterRange

        .. todo:: Don't include game as param, it's only used in outputs, which should not be included here.
        .. todo:: Add some try-except logic around the `int(...)` calls.
        .. todo:: Add logic to check for yyyymmdd in addition to mmddyyyy.
        .. todo:: Add logic to check for `-` separators, in addition to `/`.
        """
        _from : datetime
        _to   : datetime
        today : datetime = datetime.now()
        # If we want to export all data for a given month, calculate a date range from 1st to end of month.
        if monthly:
            month : int = today.month
            year  : int = today.year
            month_year = start_date.split("/")
            month = int(month_year[0])
            year  = int(month_year[1])
            month_range = monthrange(year, month)
            days_in_month = month_range[1]
            _from = datetime(year=year, month=month, day=1, hour=0, minute=0, second=0)
            _to   = datetime(year=year, month=month, day=days_in_month, hour=23, minute=59, second=59)
            Logger.Log(f"Exporting {month}/{year} data for {game}...", logging.DEBUG)
        # Otherwise, create date range from given pair of dates.
        else:
            _from = datetime.strptime(start_date, "%m/%d/%Y") if start_date is not None else today
            _from = _from.replace(hour=0, minute=0, second=0)
            _to   = datetime.strptime(end_date, "%m/%d/%Y") if end_date is not None else _from
            _to = _to.replace(hour=23, minute=59, second=59)
            if _from > _to:
                raise ValueError(f"Invalid date range, start date of {_from} is after end date of {_to}!")
            Logger.Log(f"Exporting from {str(_from)} to {str(_to)} of data for {game}...", logging.INFO)
        return ExporterRange.FromDateRange(source=interface, date_min=_from, date_max=_to)



## import standard libraries
import json
import logging
import os
import re
import shutil
import sys
import traceback
import zipfile
from datetime import datetime
from git.repo import Repo
from git.exc import InvalidGitRepositoryError, NoSuchPathError
from pathlib import Path
from typing import Any, Dict, IO, List, Optional, Set

# import local files
from ogd import games
from ogd.core.interfaces.outerfaces.DataOuterface import DataOuterface
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.schemas.tables.TableSchema import TableSchema
from ogd.core.schemas.configs.IndexingSchema import FileIndexingSchema
from ogd.core.utils import utils
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow
from ogd.core.utils.Readme import Readme

class TSVOuterface(DataOuterface):

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_id:str, config:GameSourceSchema, export_modes:Set[ExportMode], date_range:Dict[str,Optional[datetime]], file_indexing:FileIndexingSchema, extension:str="tsv", dataset_id:Optional[str]=None):
        super().__init__(game_id=game_id, config=config, export_modes=export_modes)
        self._file_paths    : Dict[str,Optional[Path]] = {"population":None, "players":None, "sessions":None, "processed_events":None, "raw_events":None}
        self._zip_paths     : Dict[str,Optional[Path]] = {"population":None, "players":None, "sessions":None, "processed_events":None, "raw_events":None}
        self._files         : Dict[str,Optional[IO]]   = {"population":None, "players":None, "sessions":None, "processed_events":None, "raw_events":None}
        self._file_indexing : FileIndexingSchema       = file_indexing
        self._data_dir      : Path = Path(f"./{self._file_indexing.LocalDirectory}")
        self._game_data_dir : Path = self._data_dir / self._game_id
        self._readme_path   : Path = self._game_data_dir / "README.md"
        self._extension     : str  = extension
        self._date_range    : Dict[str,Optional[datetime]] = date_range
        self._dataset_id    : str  = ""
        self._short_hash    : str  = ""
        # self._sess_count    : int  = 0
        # figure out dataset ID.
        # Logger.Log(f"the min in date range is {self._date_range['min']}, of type {type(self._date_range['min'])}")
        start = self._date_range['min'].strftime("%Y%m%d") if self._date_range['min'] is not None else "UNKNOWN"
        end   = self._date_range['max'].strftime("%Y%m%d") if self._date_range['max'] is not None else "UNKNOWN"
        self._dataset_id = dataset_id or f"{self._game_id}_{start}_to_{end}"
        # get hash
        try:
            repo = Repo(search_parent_directories=True)
            if repo.git is not None:
                self._short_hash = str(repo.git.rev_parse(repo.head.object.hexsha, short=7))
        except InvalidGitRepositoryError as err:
            msg = f"Code is not in a valid Git repository:\n{str(err)}"
            Logger.Log(msg, logging.ERROR)
        except NoSuchPathError as err:
            msg = f"Unable to access proper file paths for Git repository:\n{str(err)}"
            Logger.Log(msg, logging.ERROR)
        # then set up our paths, and ensure each exists.
        base_file_name    : str  = f"{self._dataset_id}_{self._short_hash}"
        # finally, generate file names.
        if ExportMode.EVENTS in export_modes:
            self._file_paths['raw_events']       = self._game_data_dir / f"{base_file_name}_events.{self._extension}"
            self._zip_paths['raw_events']        = self._game_data_dir / f"{base_file_name}_events.zip"
        if ExportMode.DETECTORS in export_modes:
            self._file_paths['processed_events'] = self._game_data_dir / f"{base_file_name}_all-events.{self._extension}"
            self._zip_paths['processed_events']  = self._game_data_dir / f"{base_file_name}_all-events.zip"
        if ExportMode.SESSION in export_modes:
            self._file_paths['sessions']         = self._game_data_dir / f"{base_file_name}_session-features.{self._extension}"
            self._zip_paths['sessions']          = self._game_data_dir / f"{base_file_name}_session-features.zip"
        if ExportMode.PLAYER in export_modes:
            self._file_paths['players']          = self._game_data_dir / f"{base_file_name}_player-features.{self._extension}"
            self._zip_paths['players']           = self._game_data_dir / f"{base_file_name}_player-features.zip"
        if ExportMode.POPULATION in export_modes:
            self._file_paths['population']       = self._game_data_dir / f"{base_file_name}_population-features.{self._extension}"
            self._zip_paths['population']        = self._game_data_dir / f"{base_file_name}_population-features.zip"
        # self.Open()

    def __del__(self):
        self.Close()

    # *** IMPLEMENT ABSTRACTS ***

    def _open(self) -> bool:
        self._game_data_dir.mkdir(exist_ok=True, parents=True)
        if (self._file_paths['raw_events'] is not None):
            self._files['raw_events']       = open(self._file_paths['raw_events'],   "w+", encoding="utf-8")
        if (self._file_paths['processed_events'] is not None):
            self._files['processed_events'] = open(self._file_paths['processed_events'],   "w+", encoding="utf-8")
        if (self._file_paths['sessions'] is not None):
            self._files['sessions']         = open(self._file_paths['sessions'], "w+", encoding="utf-8")
        if (self._file_paths['players'] is not None):
            self._files['players']          = open(self._file_paths['players'],  "w+", encoding="utf-8")           
        if (self._file_paths['population'] is not None):
            self._files['population']       = open(self._file_paths['population'], "w+", encoding="utf-8")
        return True

    def _close(self) -> bool:
        Logger.Log(f"Closing TSV outerface...")
        try:
            # before we zip stuff up, let's check if the readme is in place:
            readme = open(self._readme_path, mode='r')
        except FileNotFoundError:
            # if not in place, generate the readme
            Logger.Log(f"Missing readme for {self._game_id}, generating new readme...", logging.WARNING, depth=1)
            _games_path  = Path(games.__file__) if Path(games.__file__).is_dir() else Path(games.__file__).parent
            game_schema  : GameSchema  = GameSchema.FromFile(game_id=self._game_id, schema_path=_games_path / self._game_id / "schemas")
            table_schema = TableSchema(schema_name=self._config.TableSchema)
            readme = Readme(game_schema=game_schema, table_schema=table_schema)
            readme.GenerateReadme(path=self._game_data_dir)
        else:
            # otherwise, readme is there, so just close it and move on.
            readme.close()
            Logger.Log(f"Successfully found, opened, and closed the README.md", logging.DEBUG, depth=1)
        finally:
            self._closeFiles()
            self._zipFiles()
            self._writeMetadataFile(num_sess=self.SessionCount)
            self._updateFileExportList(num_sess=self.SessionCount)
            return True

    def _destination(self, mode:ExportMode) -> str:
        ret_val = ""
        match mode:
            case ExportMode.EVENTS:
                ret_val = str(self._file_paths['raw_events'])
            case ExportMode.DETECTORS:
                ret_val = str(self._file_paths['processed_events'])
            case ExportMode.SESSION:
                ret_val = str(self._file_paths['sessions'])
            case ExportMode.PLAYER:
                ret_val = str(self._file_paths['players'])
            case ExportMode.POPULATION:
                ret_val = str(self._file_paths['population'])
        return ret_val

    def _removeExportMode(self, mode:ExportMode):
        match mode:
            # NOTE: Originally, this case was a lone "if" followed by "if-elif-elif..." for Detectors on.
            # I don't think that was intentional, but wanted to make a note of it just in case.
            case ExportMode.EVENTS:
                if self._files['raw_events'] is not None:
                    self._files['raw_events'].close()
                self._files['raw_events']      = None
                self._file_paths['raw_events'] = None
                self._zip_paths['raw_events']  = None
            case ExportMode.DETECTORS:
                if self._files['processed_events'] is not None:
                    self._files['processed_events'].close()
                self._files['processed_events']      = None
                self._file_paths['processed_events'] = None
                self._zip_paths['processed_events']  = None
            case ExportMode.SESSION:
                if self._files['sessions'] is not None:
                    self._files['sessions'].close()
                self._files['sessions']      = None
                self._file_paths['sessions'] = None
                self._zip_paths['sessions']  = None
            case ExportMode.PLAYER:
                if self._files['players'] is not None:
                    self._files['players'].close()
                self._files['players']      = None
                self._file_paths['players'] = None
                self._zip_paths['players']  = None
            case ExportMode.POPULATION:
                if self._files['population'] is not None:
                    self._files['population'].close()
                self._files['population']      = None
                self._file_paths['population'] = None
                self._zip_paths['population']  = None

    def _writeRawEventsHeader(self, header:List[str]) -> None:
        cols = TSVOuterface._cleanSpecialChars(vals=header)
        cols_line = "\t".join(cols) + "\n"
        if self._files['raw_events'] is not None:
            self._files['raw_events'].writelines(cols_line)
        else:
            Logger.Log("No raw_events file available, writing to standard output instead.", logging.WARN)
            sys.stdout.write("".join(cols_line))

    def _writeProcessedEventsHeader(self, header:List[str]) -> None:
        cols = TSVOuterface._cleanSpecialChars(vals=header)
        cols_line = "\t".join(cols) + "\n"
        if self._files['processed_events'] is not None:
            self._files['processed_events'].writelines(cols_line)
        else:
            Logger.Log("No processed_events file available, writing to standard output instead.", logging.WARN)
            sys.stdout.write("".join(cols_line))

    def _writeSessionHeader(self, header:List[str]) -> None:
        cols = TSVOuterface._cleanSpecialChars(vals=header)
        cols_line = "\t".join(cols) + "\n"
        if self._files['sessions'] is not None:
            self._files['sessions'].writelines(cols_line)
        else:
            Logger.Log("No session file available, writing to standard output instead.", logging.WARN)
            sys.stdout.write("".join(cols_line))

    def _writePlayerHeader(self, header:List[str]) -> None:
        cols = TSVOuterface._cleanSpecialChars(vals=header)
        cols_line = "\t".join(cols) + "\n"
        if self._files['players'] is not None:
            self._files['players'].writelines(cols_line)
        else:
            Logger.Log("No player file available, writing to standard output instead.", logging.WARN)
            sys.stdout.write("".join(cols_line))

    def _writePopulationHeader(self, header:List[str]) -> None:
        cols = TSVOuterface._cleanSpecialChars(vals=header)
        cols_line = "\t".join(cols) + "\n"
        if self._files['population'] is not None:
            self._files['population'].writelines(cols_line)
        else:
            Logger.Log("No population file available, writing to standard output instead.", logging.WARN)
            sys.stdout.write("".join(cols_line))

    def _writeRawEventLines(self, events:List[ExportRow]) -> None:
        event_strs = [TSVOuterface._cleanSpecialChars(vals=[str(item) for item in event]) for event in events]
        event_lines = ["\t".join(event) + "\n" for event in event_strs]
        if self._files['raw_events'] is not None:
            self._files['raw_events'].writelines(event_lines)
        else:
            Logger.Log("No raw_events file available, writing to standard output instead.", logging.WARN)
            sys.stdout.write("".join(event_lines))

    def _writeProcessedEventLines(self, events:List[ExportRow]) -> None:
        event_strs = [TSVOuterface._cleanSpecialChars(vals=[str(item) for item in event]) for event in events]
        event_lines = ["\t".join(event) + "\n" for event in event_strs]
        if self._files['processed_events'] is not None:
            self._files['processed_events'].writelines(event_lines)
        else:
            Logger.Log("No processed_events file available, writing to standard output instead.", logging.WARN)
            sys.stdout.write("".join(event_lines))

    def _writeSessionLines(self, sessions:List[ExportRow]) -> None:
        # self._sess_count += len(sessions)
        _session_feats = [TSVOuterface._cleanSpecialChars(vals=sess) for sess in sessions]
        _session_lines = ["\t".join(sess) + "\n" for sess in _session_feats]
        if self._files['sessions'] is not None:
            self._files['sessions'].writelines(_session_lines)
        else:
            Logger.Log("No session file available, writing to standard output instead.", logging.WARN)
            sys.stdout.write("".join(_session_lines))

    def _writePlayerLines(self, players:List[ExportRow]) -> None:
        _player_feats = [TSVOuterface._cleanSpecialChars(vals=play) for play in players]
        _player_lines = ["\t".join(play) + "\n" for play in _player_feats]
        if self._files['players'] is not None:
            self._files['players'].writelines(_player_lines)
        else:
            Logger.Log("No player file available, writing to standard output instead.", logging.WARN)
            sys.stdout.write("".join(_player_lines))

    def _writePopulationLines(self, populations:List[ExportRow]) -> None:
        _pop_feats = [TSVOuterface._cleanSpecialChars(vals=pop) for pop in populations]
        _pop_lines = ["\t".join(pop) + "\n" for pop in _pop_feats]
        if self._files['population'] is not None:
            self._files['population'].writelines(_pop_lines)
        else:
            Logger.Log("No population file available, writing to standard output instead.", logging.WARN)
            sys.stdout.write("".join(_pop_lines))

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    @staticmethod
    def _cleanSpecialChars(vals:List[Any], tab_width:int=3) -> List[str]:
        # check all return values for strings, and ensure no newlines or tabs get through, as they could throw off our outputs.
        for i in range(len(vals)):
            vals[i] = str(vals[i]).replace('\n', ' ').replace('\t', ' '*tab_width)
        return vals

    @staticmethod
    def _addToZip(path, zip_file, path_in_zip) -> None:
        try:
            zip_file.write(path, path_in_zip)
        except FileNotFoundError as err:
            Logger.Log(str(err), logging.ERROR)
            traceback.print_tb(err.__traceback__)

    # *** PRIVATE METHODS ***

    def _closeFiles(self) -> None:
        if self._files['raw_events'] is not None:
            self._files['raw_events'].close()
        if self._files['processed_events'] is not None:
            self._files['processed_events'].close()
        if self._files['sessions'] is not None:
            self._files['sessions'].close()
        if self._files['players'] is not None:
            self._files['players'].close()
        if self._files['population'] is not None:
            self._files['population'].close()

    def _zipFiles(self) -> None:
        existing_datasets = {}
        try:
            file_directory = utils.loadJSONFile(filename="file_list.json", path=self._data_dir)
            existing_datasets = file_directory.get(self._game_id, {})
        except FileNotFoundError as err:
            Logger.Log("file_list.json does not exist.", logging.WARNING)
        except json.decoder.JSONDecodeError as err:
            Logger.Log(f"file_list.json has invalid format: {str(err)}.", logging.WARNING)
        # if we have already done this dataset before, rename old zip files
        # (of course, first check if we ever exported this game before).
        existing_meta = existing_datasets.get(self._dataset_id, None)
        if existing_meta is not None:
            _existing_raw_events_file  = existing_meta.get('raw_events_file', None)
            _existing_proc_events_file = existing_meta.get('processed_events_file', None) or existing_meta.get('events_file', None)
            _existing_sess_file    = existing_meta.get('sessions_file', None)
            _existing_players_file = existing_meta.get('players_file', None)
            _existing_pop_file     = existing_meta.get('population_file', None)
            try:
                if _existing_raw_events_file is not None and Path(_existing_raw_events_file).is_file() and self._zip_paths['raw_events'] is not None:
                    Logger.Log(f"Renaming {str(_existing_raw_events_file)} -> {self._zip_paths['raw_events']}", logging.DEBUG)
                    os.rename(_existing_raw_events_file, str(self._zip_paths['raw_events']))
                if _existing_proc_events_file is not None and Path(_existing_proc_events_file).is_file() and self._zip_paths['processed_events'] is not None:
                    Logger.Log(f"Renaming {str(_existing_proc_events_file)} -> {self._zip_paths['processed_events']}", logging.DEBUG)
                    os.rename(_existing_proc_events_file, str(self._zip_paths['processed_events']))
                if _existing_sess_file is not None and Path(_existing_sess_file).is_file() and self._zip_paths['sessions'] is not None:
                    Logger.Log(f"Renaming {str(_existing_sess_file)} -> {self._zip_paths['sessions']}", logging.DEBUG)
                    os.rename(_existing_sess_file, str(self._zip_paths['sessions']))
                if _existing_players_file is not None and Path(_existing_players_file).is_file() and self._zip_paths['players'] is not None:
                    Logger.Log(f"Renaming {str(_existing_players_file)} -> {self._zip_paths['players']}", logging.DEBUG)
                    os.rename(_existing_players_file, str(self._zip_paths['players']))
                if _existing_pop_file is not None and Path(_existing_pop_file).is_file() and self._zip_paths['population'] is not None:
                    Logger.Log(f"Renaming {str(_existing_pop_file)} -> {self._zip_paths['population']}", logging.DEBUG)
                    os.rename(_existing_pop_file, str(self._zip_paths['population']))
            except FileExistsError as err:
                msg = f"Error while setting up zip files, could not rename an existing file because another file is already using the target name! {err}"
                Logger.Log(msg, logging.ERROR)
            except Exception as err:
                msg = f"Unexpected error while setting up zip files! {type(err)} : {err}"
                Logger.Log(msg, logging.ERROR)
                traceback.print_tb(err.__traceback__)
        # for each file, try to save out the csv/tsv to a file - if it's one that should be exported, that is.
        if self._zip_paths['population'] is not None:
            with zipfile.ZipFile(self._zip_paths["population"], "w", compression=zipfile.ZIP_DEFLATED) as population_zip_file:
                try:
                    population_file = Path(self._dataset_id) / f"{self._dataset_id}_{self._short_hash}_population-features.{self._extension}"
                    readme_file  = Path(self._dataset_id) / "README.md"
                    TSVOuterface._addToZip(path=self._file_paths["population"], zip_file=population_zip_file, path_in_zip=population_file)
                    TSVOuterface._addToZip(path=self._readme_path,              zip_file=population_zip_file, path_in_zip=readme_file)
                    population_zip_file.close()
                    if self._file_paths["population"] is not None:
                        os.remove(self._file_paths["population"])
                except FileNotFoundError as err:
                    Logger.Log(f"FileNotFoundError Exception: {err}", logging.ERROR)
                    traceback.print_tb(err.__traceback__)
        if self._zip_paths['players'] is not None:
            with zipfile.ZipFile(self._zip_paths["players"], "w", compression=zipfile.ZIP_DEFLATED) as players_zip_file:
                try:
                    player_file = Path(self._dataset_id) / f"{self._dataset_id}_{self._short_hash}_player-features.{self._extension}"
                    readme_file  = Path(self._dataset_id) / "README.md"
                    TSVOuterface._addToZip(path=self._file_paths["players"], zip_file=players_zip_file, path_in_zip=player_file)
                    TSVOuterface._addToZip(path=self._readme_path,           zip_file=players_zip_file, path_in_zip=readme_file)
                    players_zip_file.close()
                    if self._file_paths["players"] is not None:
                        os.remove(self._file_paths["players"])
                except FileNotFoundError as err:
                    Logger.Log(f"FileNotFoundError Exception: {err}", logging.ERROR)
                    traceback.print_tb(err.__traceback__)
        if self._zip_paths['sessions'] is not None:
            with zipfile.ZipFile(self._zip_paths["sessions"], "w", compression=zipfile.ZIP_DEFLATED) as sessions_zip_file:
                try:
                    session_file = Path(self._dataset_id) / f"{self._dataset_id}_{self._short_hash}_session-features.{self._extension}"
                    readme_file  = Path(self._dataset_id) / "README.md"
                    TSVOuterface._addToZip(path=self._file_paths["sessions"], zip_file=sessions_zip_file, path_in_zip=session_file)
                    TSVOuterface._addToZip(path=self._readme_path,            zip_file=sessions_zip_file, path_in_zip=readme_file)
                    sessions_zip_file.close()
                    if self._file_paths["sessions"] is not None:
                        os.remove(self._file_paths["sessions"])
                except FileNotFoundError as err:
                    Logger.Log(f"FileNotFoundError Exception: {err}", logging.ERROR)
                    traceback.print_tb(err.__traceback__)
        if self._zip_paths['raw_events'] is not None:
            with zipfile.ZipFile(self._zip_paths["raw_events"], "w", compression=zipfile.ZIP_DEFLATED) as _raw_events_zip_file:
                try:
                    events_file = Path(self._dataset_id) / f"{self._dataset_id}_{self._short_hash}_events.{self._extension}"
                    readme_file = Path(self._dataset_id) / "README.md"
                    TSVOuterface._addToZip(path=self._file_paths["raw_events"], zip_file=_raw_events_zip_file, path_in_zip=events_file)
                    TSVOuterface._addToZip(path=self._readme_path,          zip_file=_raw_events_zip_file, path_in_zip=readme_file)
                    _raw_events_zip_file.close()
                    if self._file_paths["raw_events"] is not None:
                        os.remove(self._file_paths["raw_events"])
                except FileNotFoundError as err:
                    Logger.Log(f"FileNotFoundError Exception: {err}", logging.ERROR)
                    traceback.print_tb(err.__traceback__)
        if self._zip_paths['processed_events'] is not None:
            with zipfile.ZipFile(self._zip_paths["processed_events"], "w", compression=zipfile.ZIP_DEFLATED) as _processed_events_zip_file:
                try:
                    events_file = Path(self._dataset_id) / f"{self._dataset_id}_{self._short_hash}_all-events.{self._extension}"
                    readme_file = Path(self._dataset_id) / "README.md"
                    TSVOuterface._addToZip(path=self._file_paths["processed_events"], zip_file=_processed_events_zip_file, path_in_zip=events_file)
                    TSVOuterface._addToZip(path=self._readme_path,          zip_file=_processed_events_zip_file, path_in_zip=readme_file)
                    _processed_events_zip_file.close()
                    if self._file_paths["processed_events"] is not None:
                        os.remove(self._file_paths["processed_events"])
                except FileNotFoundError as err:
                    Logger.Log(f"FileNotFoundError Exception: {err}", logging.ERROR)
                    traceback.print_tb(err.__traceback__)

    ## Public function to write out a tiny metadata file for indexing OGD data files.
    #  Using the paths of the exported files, and given some other variables for
    #  deriving file metadata, this simply outputs a new file_name.meta file.
    #  @param date_range    The range of dates included in the exported data.
    #  @param num_sess      The number of sessions included in the recent export.
    def _writeMetadataFile(self, num_sess:int) -> None:
        # First, ensure we have a data directory.
        try:
            self._game_data_dir.mkdir(exist_ok=True, parents=True)
        except Exception as err:
            msg = f"Could not set up folder {self._game_data_dir}. {type(err)} {str(err)}"
            Logger.Log(msg, logging.WARNING)
        else:
            # Second, remove old metas, if they exist.
            start_range = self._date_range['min'].strftime("%Y%m%d") if self._date_range['min'] is not None else "Unknown"
            end_range   = self._date_range['max'].strftime("%Y%m%d") if self._date_range['max'] is not None else "Unknown"
            match_string = f"{self._game_id}_{start_range}_to_{end_range}_\\w*\\.meta"
            old_metas = [f for f in os.listdir(self._game_data_dir) if re.match(match_string, f)]
            for old_meta in old_metas:
                try:
                    Logger.Log(f"Removing old meta file, {old_meta}")
                    os.remove(self._game_data_dir / old_meta)
                except Exception as err:
                    msg = f"Could not remove old meta file {old_meta}. {type(err)} {str(err)}"
                    Logger.Log(msg, logging.WARNING)
            # Third, write the new meta file.
            # calculate the path and name of the metadata file, and open/make it.
            meta_file_path : Path = self._game_data_dir/ f"{self._dataset_id}_{self._short_hash}.meta"
            with open(meta_file_path, "w", encoding="utf-8") as meta_file :
                metadata  = \
                {
                    "game_id"      :self._game_id,
                    "dataset_id"   :self._dataset_id,
                    "ogd_revision" :self._short_hash,
                    "start_date"   :self._date_range['min'].strftime("%m/%d/%Y") if self._date_range['min'] is not None else "Unknown",
                    "end_date"     :self._date_range['max'].strftime("%m/%d/%Y") if self._date_range['max'] is not None else "Unknown",
                    "date_modified":datetime.now().strftime("%m/%d/%Y"),
                    "sessions"     :num_sess,
                    "population_file"     : str(self._zip_paths['population'])       if self._zip_paths['population']       else None,
                    "population_template" : f'/tree/{self._game_id.lower()}'         if self._zip_paths['population']       else None,
                    "players_file"        : str(self._zip_paths['players'])          if self._zip_paths['players']          else None,
                    "players_template"    : f'/tree/{self._game_id.lower()}'         if self._zip_paths['players']          else None,
                    "sessions_file"       : str(self._zip_paths['sessions'])         if self._zip_paths['sessions']         else None,
                    "sessions_template"   : f'/tree/{self._game_id.lower()}'         if self._zip_paths['sessions']         else None,
                    "raw_file"            : str(self._zip_paths['raw_events'])       if self._zip_paths['raw_events']       else None,
                    "events_template"     : f'/tree/{self._game_id.lower()}'         if self._zip_paths['raw_events']       else None,
                    "events_file"         : str(self._zip_paths['processed_events']) if self._zip_paths['processed_events'] else None,
                    "all_events_template" : f'/tree/{self._game_id.lower()}'         if self._zip_paths['processed_events'] else None
                }
                meta_file.write(json.dumps(metadata, indent=4))
                meta_file.close()

    ## Public function to update the list of exported files.
    #  Using the paths of the exported files, and given some other variables for
    #  deriving file metadata, this simply updates the JSON file to the latest
    #  list of files.
    #  @param date_range    The range of dates included in the exported data.
    #  @param num_sess      The number of sessions included in the recent export.
    def _updateFileExportList(self, num_sess: int) -> None:
        self._backupFileExportList()
        file_index = {}
        existing_datasets = {}
        try:
            file_index = utils.loadJSONFile(filename="file_list.json", path=self._data_dir)
        except FileNotFoundError as err:
            Logger.Log("file_list.json does not exist.", logging.WARNING)
        except json.decoder.JSONDecodeError as err:
            Logger.Log(f"file_list.json has invalid format: {str(err)}.", logging.WARNING)
        finally:
            if not "CONFIG" in file_index.keys():
                Logger.Log(f"No CONFIG found in file_list.json, adding default CONFIG...", logging.WARNING)
                file_index["CONFIG"] = {
                    "files_base" : self._file_indexing.RemoteURL,
                    "templates_base" : self._file_indexing.TemplatesURL
                }
            if not self._game_id in file_index.keys():
                file_index[self._game_id] = {}
            existing_datasets  = file_index[self._game_id]
            with open(self._data_dir / "file_list.json", "w") as existing_csv_file:
                Logger.Log(f"Opened file list for writing at {existing_csv_file.name}", logging.INFO)
                existing_metadata = existing_datasets.get(self._dataset_id, {})
                population_path = self._zip_paths.get("population") or existing_metadata.get("population")
                players_path    = self._zip_paths.get("players")    or existing_metadata.get("players")
                sessions_path   = self._zip_paths.get("sessions")   or existing_metadata.get("sessions")
                raw_events_path = self._zip_paths.get("raw_events") or existing_metadata.get("raw_events") or existing_metadata.get("events")
                processed_events_path = self._zip_paths.get("processed_events") or existing_metadata.get("processed_events")
                file_index[self._game_id][self._dataset_id] = \
                {
                    "ogd_revision"          : self._short_hash,
                    "start_date"            : self._date_range['min'].strftime("%m/%d/%Y") if self._date_range['min'] is not None else "Unknown",
                    "end_date"              : self._date_range['max'].strftime("%m/%d/%Y") if self._date_range['max'] is not None else "Unknown",
                    "date_modified"         : datetime.now().strftime("%m/%d/%Y"),
                    "sessions"              : num_sess,
                    "population_file"       : str(population_path)       if population_path       is not None else None,
                    "population_template"   : ''                         if population_path       is not None else None,
                    "players_file"          : str(players_path)          if players_path          is not None else None,
                    "players_template"      : ''                         if players_path          is not None else None,
                    "sessions_file"         : str(sessions_path)         if sessions_path         is not None else None,
                    "sessions_template"     : ''                         if sessions_path         is not None else None,
                    "events_file"           : str(raw_events_path)       if raw_events_path       is not None else None,
                    "events_template"       : ''                         if raw_events_path       is not None else None,
                    "all_events_file"       : str(processed_events_path) if processed_events_path is not None else None,
                    "all_events_template"   : ''                         if processed_events_path is not None else None
                }
                existing_csv_file.write(json.dumps(file_index, indent=4))

    def _backupFileExportList(self) -> bool:
        try:
            src  : Path = self._data_dir / "file_list.json"
            dest : Path = self._data_dir / "file_list.json.bak"
            if src.exists():
                shutil.copyfile(src=src, dst=dest)
            else:
                Logger.Log(f"Could not back up file_list.json, because it does not exist!", logging.WARN)
        except Exception as err:
            msg = f"{type(err)} {str(err)}"
            Logger.Log(f"Could not back up file_list.json. Got the following error: {msg}", logging.ERROR)
            return False
        else:
            Logger.Log(f"Backed up file_list.json to {dest}", logging.INFO)
            return True


# import libraries
import logging
from datetime import datetime, timedelta
from typing import Any, List
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.utils.Logger import Logger

class SessionDuration(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters, session_id:str):
        self._session_id = session_id
        super().__init__(params=params)
        self._start_time = None
        self._start_index = None
        self._latest_event = None
        # self._session_duration = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        # if this was earliest event, make it the start time.
        if not self._start_time:
            self._start_time = event.Timestamp
            self._start_index = event.EventSequenceIndex
        if self._start_time > event.Timestamp + timedelta(milliseconds=100):
            # Logger.Log(f"Got out-of-order events in SessionDuration; event {event.EventName}:{event.EventSequenceIndex} for player {event.UserID}:{event.SessionID} had timestamp {event.Timestamp} earlier than start event, with time {self._start_time}, index {self._start_index}!", logging.WARN)
            self._start_time = event.Timestamp
            self._start_index = event.EventSequenceIndex
        # if this was the latest event, make it the end time, otherwise output error.
        if self._latest_event is not None and self._latest_event.Timestamp > event.Timestamp + timedelta(milliseconds=100):
            # Logger.Log(f"Got out-of-order events in SessionDuration:\n   Event {event.EventName}:{event.EventSequenceIndex} for player {event.UserID}:{event.SessionID} had timestamp {event.Timestamp},\n   Earlier than previous latest event {self._latest_event.EventName}:{self._latest_event.EventSequenceIndex} for player {self._latest_event.UserID}:{self._latest_event.SessionID} with timestamp {self._latest_event.Timestamp}", logging.WARN)
            pass
        else:
            self._latest_event = event
        # self._session_duration = (event.Timestamp - self._client_start_time).total_seconds()

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if self._start_time and self._latest_event:
            return [self._latest_event.Timestamp - self._start_time]
        else:
            return ["No events"]

    # *** Optionally override public functions. ***

    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.SESSION]


# import libraries
from typing import Any, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class GameCompletionStatus(Feature):
    def __init__(self, params: GeneratorParameters):
        super().__init__(params=params)
        self.status: Optional[str] = "IN_PROGRESS"

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode: ExtractionMode) -> List[str]:
        return ["win_game", "lose_game"]

    @classmethod
    def _featureFilter(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        if event.EventName == "win_game":
            self.status = "WIN"
        elif event.EventName == "lose_game":
            self.status = "LOSS"

    def _updateFromFeatureData(self, feature: FeatureData):
        pass

    def _getFeatureValues(self) -> List[Any]:
        return [self.status]



# import standard libraries
import abc
import logging
from typing import Any, Dict, List, Optional
# import local files
from ogd.core.utils.Logger import Logger

class Schema(abc.ABC):
    def __init__(self, name:str, other_elements:Optional[Dict[str, Any]]):
        self._name : str
        self._other_elements : Dict[str, Any]

        self._name = Schema._parseName(name)

        self._other_elements = other_elements or {}
        if len(self._other_elements.keys()) > 0:
            Logger.Log(f"Schema for {self.Name} contained nonstandard elements {self.NonStandardElementNames}")

    def __str__(self):
        return f"{type(self).__name__}[{self.Name}]"

    def __repr__(self):
        return f"{type(self).__name__}[{self.Name}]"

    @property
    @abc.abstractmethod
    def AsMarkdown(self) -> str:
        pass

    @property
    def Name(self) -> str:
        return self._name

    @property
    def NonStandardElements(self) -> Dict[str, Any]:
        return self._other_elements

    @property
    def NonStandardElementNames(self) -> List[str]:
        return list(self._other_elements.keys())
    
    @staticmethod
    def _parseName(name):
        ret_val : str
        if isinstance(name, str):
            ret_val = name
        else:
            ret_val = str(name)
            Logger.Log(f"Schema name was not a string, defaulting to str(name) == {ret_val}", logging.WARN)
        return ret_val